/*
 *
 *  BlueZ - Bluetooth protocol stack for Linux
 *
 *  Copyright (C) 2006-2007  Nokia Corporation
 *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define _GNU_SOURCE
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/ioctl.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/l2cap.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>

#include <glib.h>
#include <dbus/dbus.h>
#include <gdbus.h>

#include "logging.h"
#include "textfile.h"

#include "hcid.h"
#include "sdpd.h"
#include "sdp-xml.h"
#include "manager.h"
#include "adapter.h"
#include "device.h"
#include "dbus-common.h"
#include "dbus-hci.h"
#include "error.h"
#include "glib-helper.h"
#include "agent.h"
#include "storage.h"

#include "fmradio_bcm4325.h"

//for debug
#define LOG_TAG "BT_HCID"
#define LOG_NIDEBUG 0
#include <utils/Log.h>

//20090924 RDS
static uint8_t  RDSGroupCode = 0xFF;
static uint8_t  RDSInvalidBlockB = 1;
static uint8_t  RDSDISegmentAddr = 0;
static uint16_t RDSPICode  = 0;
static uint16_t RDSPrevPI = 0;
static uint16_t RDSCurrentPI = 0;
static uint8_t  RDSPIMatched = 1;
static uint8_t  RDSPrevBlockID = 0;
static uint8_t  RDSNextBlockID = 0;
static uint8_t  RDSProgramServiceName[BCM4325_RDS_PSN_LENGTH +1] ={ ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', '\0' }; 
#define NUM_ELEMENTS(table) (sizeof(table)/sizeof(const char *))

#define IO_CAPABILITY_DISPLAYONLY	0x00
#define IO_CAPABILITY_DISPLAYYESNO	0x01
#define IO_CAPABILITY_KEYBOARDONLY	0x02
#define IO_CAPABILITY_NOINPUTNOOUTPUT	0x03
#define IO_CAPABILITY_INVALID		0xFF

#define check_address(address) bachk(address)

static DBusConnection *connection = NULL;
static GSList *adapter_drivers = NULL;

const struct btd_adapter_ops *adapter_ops = NULL;

struct session_req {
	struct btd_adapter	*adapter;
	DBusConnection		*conn;		/* Connection reference */
	DBusMessage		*msg;		/* Unreplied message ref */
	char			*owner;		/* Bus name of the owner */
	guint			id;		/* Listener id */
	uint8_t			mode;		/* Requested mode */
	int			refcount;	/* Session refcount */
};

struct service_auth {
	service_auth_cb cb;
	void *user_data;
	struct btd_device *device;
};

struct btd_adapter {
	uint16_t dev_id;
	int up;
	char *path;			/* adapter object path */
	bdaddr_t bdaddr;		/* adapter Bluetooth Address */
	guint discov_timeout_id;	/* discoverable timeout id */
	uint32_t discov_timeout;	/* discoverable time(sec) */
	guint pairable_timeout_id;	/* pairable timeout id */
	uint32_t pairable_timeout;	/* pairable time(sec) */
	uint8_t scan_mode;		/* scan mode: SCAN_DISABLED, SCAN_PAGE,
					 * SCAN_INQUIRY */
	uint8_t mode;			/* off, connectable, discoverable,
					 * limited */
	uint8_t global_mode;		/* last valid global mode */
	int state;			/* standard inq, periodic inq, name
					 * resloving */
	GSList *found_devices;
	GSList *oor_devices;		/* out of range device list */
	DBusMessage *discovery_cancel;	/* discovery cancel message request */
	GSList *passkey_agents;
	struct agent *agent;		/* For the new API */
	GSList *connections;		/* Connected devices */
	GSList *devices;		/* Devices structure pointers */
	GSList *mode_sessions;		/* Request Mode sessions */
	GSList *disc_sessions;		/* Discovery sessions */
	guint scheduler_id;		/* Scheduler handle */

	struct hci_dev dev;		/* hci info */
	gboolean pairable;		/* pairable state */

	gboolean initialized;
	gboolean already_up;		/* adapter was already up on init */

	gboolean off_requested;		/* DEVDOWN ioctl was called */

	uint8_t svc_cache;		/* Service Class cache */
	gboolean cache_enable;

	gint ref;
};

static void adapter_set_pairable_timeout(struct btd_adapter *adapter,
					guint interval);

static inline DBusMessage *invalid_args(DBusMessage *msg)
{
	return g_dbus_create_error(msg, ERROR_INTERFACE ".InvalidArguments",
			"Invalid arguments in method call");
}

static inline DBusMessage *not_available(DBusMessage *msg)
{
	return g_dbus_create_error(msg, ERROR_INTERFACE ".NotAvailable",
			"Not Available");
}

static inline DBusMessage *adapter_not_ready(DBusMessage *msg)
{
	return g_dbus_create_error(msg, ERROR_INTERFACE ".NotReady",
			"Adapter is not ready");
}

static inline DBusMessage *no_such_adapter(DBusMessage *msg)
{
	return g_dbus_create_error(msg, ERROR_INTERFACE ".NoSuchAdapter",
							"No such adapter");
}

static inline DBusMessage *failed_strerror(DBusMessage *msg, int err)
{
	return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
							strerror(err));
}

static inline DBusMessage *in_progress(DBusMessage *msg, const char *str)
{
	return g_dbus_create_error(msg, ERROR_INTERFACE ".InProgress", str);
}

static inline DBusMessage *not_in_progress(DBusMessage *msg, const char *str)
{
	return g_dbus_create_error(msg, ERROR_INTERFACE ".NotInProgress", str);
}

static inline DBusMessage *not_authorized(DBusMessage *msg)
{
	return g_dbus_create_error(msg, ERROR_INTERFACE ".NotAuthorized",
			"Not authorized");
}

static inline DBusMessage *unsupported_major_class(DBusMessage *msg)
{
	return g_dbus_create_error(msg,
			ERROR_INTERFACE ".UnsupportedMajorClass",
			"Unsupported Major Class");
}

static int found_device_cmp(const struct remote_dev_info *d1,
			const struct remote_dev_info *d2)
{
	int ret;

	if (bacmp(&d2->bdaddr, BDADDR_ANY)) {
		ret = bacmp(&d1->bdaddr, &d2->bdaddr);
		if (ret)
			return ret;
	}

	if (d2->name_status != NAME_ANY) {
		ret = (d1->name_status - d2->name_status);
		if (ret)
			return ret;
	}

	return 0;
}

static void dev_info_free(struct remote_dev_info *dev)
{
	g_free(dev->name);
	g_free(dev->alias);
	g_free(dev);
}

void clear_found_devices_list(struct btd_adapter *adapter)
{
	if (!adapter->found_devices)
		return;

	g_slist_foreach(adapter->found_devices, (GFunc) dev_info_free, NULL);
	g_slist_free(adapter->found_devices);
	adapter->found_devices = NULL;
}

static int set_service_classes(struct btd_adapter *adapter, uint8_t value)
{
	struct hci_dev *dev = &adapter->dev;
	const uint8_t *cls = dev->class;
	uint32_t dev_class;
	int dd, err;

	if (cls[2] == value)
		return 0; /* Already set */

	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0) {
		err = -errno;
		error("Can't open device hci%d: %s (%d)",
				adapter->dev_id, strerror(errno), errno);
		return err;
	}

	dev_class = (value << 16) | (cls[1] << 8) | cls[0];

	debug("Changing service classes to 0x%06x", dev_class);

	if (hci_write_class_of_dev(dd, dev_class, HCI_REQ_TIMEOUT) < 0) {
		err = -errno;
		error("Can't write class of device: %s (%d)",
						strerror(errno), errno);
		hci_close_dev(dd);
		return err;
	}

	hci_close_dev(dd);

	return 0;
}

int set_major_and_minor_class(struct btd_adapter *adapter, uint8_t major,
								uint8_t minor)
{
	struct hci_dev *dev = &adapter->dev;
	const uint8_t *cls = dev->class;
	uint32_t dev_class;
	int dd, err;

	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0) {
		err = -errno;
		error("Can't open device hci%d: %s (%d)",
				adapter->dev_id, strerror(errno), errno);
		return err;
	}

	dev_class = (cls[2] << 16) | ((cls[1] & 0x20) << 8) |
						((major & 0xdf) << 8) | minor;

	debug("Changing major/minor class to 0x%06x", dev_class);

	if (hci_write_class_of_dev(dd, dev_class, HCI_REQ_TIMEOUT) < 0) {
		int err = -errno;
		error("Can't write class of device: %s (%d)",
						strerror(errno), errno);
		hci_close_dev(dd);
		return err;
	}

	hci_close_dev(dd);
	return 0;
}

int pending_remote_name_cancel(struct btd_adapter *adapter)
{
	struct remote_dev_info *dev, match;
	int err = 0;

	/* find the pending remote name request */
	memset(&match, 0, sizeof(struct remote_dev_info));
	bacpy(&match.bdaddr, BDADDR_ANY);
	match.name_status = NAME_REQUESTED;

	dev = adapter_search_found_devices(adapter, &match);
	if (!dev) /* no pending request */
		return -ENODATA;

	err = adapter_ops->cancel_resolve_name(adapter->dev_id, &dev->bdaddr);
	if (err < 0)
		error("Remote name cancel failed: %s(%d)",
						strerror(errno), errno);
	return err;
}

int adapter_resolve_names(struct btd_adapter *adapter)
{
	struct remote_dev_info *dev, match;
	int err;

	memset(&match, 0, sizeof(struct remote_dev_info));
	bacpy(&match.bdaddr, BDADDR_ANY);
	match.name_status = NAME_REQUIRED;

	dev = adapter_search_found_devices(adapter, &match);
	if (!dev)
		return -ENODATA;

	/* send at least one request or return failed if the list is empty */
	do {
		/* flag to indicate the current remote name requested */
		dev->name_status = NAME_REQUESTED;

		err = adapter_ops->resolve_name(adapter->dev_id, &dev->bdaddr);

		if (!err)
			break;

		error("Unable to send HCI remote name req: %s (%d)",
						strerror(errno), errno);

		/* if failed, request the next element */
		/* remove the element from the list */
		adapter_remove_found_device(adapter, &dev->bdaddr);

		/* get the next element */
		dev = adapter_search_found_devices(adapter, &match);
	} while (dev);

	return err;
}

static const char *mode2str(uint8_t mode)
{
	switch(mode) {
	case MODE_OFF:
		return "off";
	case MODE_CONNECTABLE:
		return "connectable";
	case MODE_DISCOVERABLE:
	case MODE_LIMITED:
		return "discoverable";
	default:
		return "unknown";
	}
}

static uint8_t get_mode(const bdaddr_t *bdaddr, const char *mode)
{
	if (strcasecmp("off", mode) == 0)
		return MODE_OFF;
	else if (strcasecmp("connectable", mode) == 0)
		return MODE_CONNECTABLE;
	else if (strcasecmp("discoverable", mode) == 0)
		return MODE_DISCOVERABLE;
	else if (strcasecmp("limited", mode) == 0)
		return MODE_LIMITED;
	else if (strcasecmp("on", mode) == 0) {
		char onmode[14], srcaddr[18];

		ba2str(bdaddr, srcaddr);
		if (read_on_mode(srcaddr, onmode, sizeof(onmode)) < 0)
			return MODE_CONNECTABLE;

		return get_mode(bdaddr, onmode);
	} else
		return MODE_UNKNOWN;
}

static void adapter_remove_discov_timeout(struct btd_adapter *adapter)
{
	if (!adapter)
		return;

	if(adapter->discov_timeout_id == 0)
		return;

	g_source_remove(adapter->discov_timeout_id);
	adapter->discov_timeout_id = 0;
}

static gboolean discov_timeout_handler(gpointer user_data)
{
	struct btd_adapter *adapter = user_data;

	adapter->discov_timeout_id = 0;

	adapter_ops->set_connectable(adapter->dev_id);

	return FALSE;
}

static void adapter_set_discov_timeout(struct btd_adapter *adapter,
					guint interval)
{
	if (adapter->discov_timeout_id) {
		g_source_remove(adapter->discov_timeout_id);
		adapter->discov_timeout_id = 0;
	}

	if (interval == 0)
		return;

	adapter->discov_timeout_id = g_timeout_add_seconds(interval,
							discov_timeout_handler,
							adapter);
}

static int set_mode(struct btd_adapter *adapter, uint8_t new_mode)
{
	int err;
	const char *modestr;

	if (!adapter->up && new_mode != MODE_OFF) {
		err = adapter_ops->set_powered(adapter->dev_id, TRUE);
		if (err < 0)
			return err;
	}

	if (adapter->up && new_mode == MODE_OFF) {
		err = adapter_ops->set_powered(adapter->dev_id, FALSE);
		if (err < 0)
			return err;

		adapter->off_requested = TRUE;

		goto done;
	}

	if (new_mode == adapter->mode)
		return 0;

	if (new_mode == MODE_CONNECTABLE)
		err = adapter_ops->set_connectable(adapter->dev_id);
	else
		err = adapter_ops->set_discoverable(adapter->dev_id);

	if (err < 0)
		return err;

	if (new_mode > MODE_CONNECTABLE) {
		adapter_remove_discov_timeout(adapter);

		if (adapter->discov_timeout)
			adapter_set_discov_timeout(adapter,
						adapter->discov_timeout);

		if (new_mode != MODE_LIMITED && adapter->mode == MODE_LIMITED)
			adapter_ops->set_limited_discoverable(adapter->dev_id,
						adapter->dev.class, FALSE);
	}

done:
	modestr = mode2str(new_mode);

	write_device_mode(&adapter->bdaddr, modestr);

	adapter->mode = new_mode;

	return 0;
}

static DBusMessage *set_powered(DBusConnection *conn, DBusMessage *msg,
				gboolean powered, void *data)
{
	struct btd_adapter *adapter = data;
	uint8_t mode;
	int err;

	mode = powered ? get_mode(&adapter->bdaddr, "on") : MODE_OFF;

	if (mode == adapter->mode)
		return dbus_message_new_method_return(msg);

	err = set_mode(adapter, mode);
	if (err < 0)
		return failed_strerror(msg, -err);

	return dbus_message_new_method_return(msg);
}

static DBusMessage *set_discoverable(DBusConnection *conn, DBusMessage *msg,
				gboolean discoverable, void *data)
{
	struct btd_adapter *adapter = data;
	uint8_t mode;
	int err;

	mode = discoverable ? MODE_DISCOVERABLE : MODE_CONNECTABLE;

	if (mode == MODE_DISCOVERABLE && adapter->pairable &&
					adapter->discov_timeout > 0 &&
					adapter->discov_timeout <= 60)
		mode = MODE_LIMITED;

	if (mode == adapter->mode)
		return dbus_message_new_method_return(msg);

	err = set_mode(adapter, mode);
	if (err < 0)
		return failed_strerror(msg, -err);

	return dbus_message_new_method_return(msg);
}

static DBusMessage *set_pairable(DBusConnection *conn, DBusMessage *msg,
				gboolean pairable, void *data)
{
	struct btd_adapter *adapter = data;
	uint8_t mode;
	int err;

	if (adapter->scan_mode == SCAN_DISABLED)
		return adapter_not_ready(msg);

	if (pairable == adapter->pairable)
		goto done;

	adapter->pairable = pairable;

	write_device_pairable(&adapter->bdaddr, pairable);

	emit_property_changed(connection, adapter->path,
				ADAPTER_INTERFACE, "Pairable",
				DBUS_TYPE_BOOLEAN, &pairable);

	if (pairable && adapter->pairable_timeout)
		adapter_set_pairable_timeout(adapter,
						adapter->pairable_timeout);

	if (!(adapter->scan_mode & SCAN_INQUIRY))
		goto done;

	mode = (pairable && adapter->discov_timeout > 0 &&
				adapter->discov_timeout <= 60) ?
					MODE_LIMITED : MODE_DISCOVERABLE;

	err = set_mode(adapter, mode);
	if (err < 0 && msg)
		return failed_strerror(msg, -err);

done:
	return msg ? dbus_message_new_method_return(msg) : NULL;
}

static gboolean pairable_timeout_handler(void *data)
{
	set_pairable(NULL, NULL, FALSE, data);

	return FALSE;
}

static void adapter_set_pairable_timeout(struct btd_adapter *adapter,
					guint interval)
{
	if (adapter->pairable_timeout_id) {
		g_source_remove(adapter->pairable_timeout_id);
		adapter->pairable_timeout_id = 0;
	}

	if (interval == 0)
		return;

	adapter->pairable_timeout_id = g_timeout_add_seconds(interval,
						pairable_timeout_handler,
						adapter);
}

static struct session_req *find_session(GSList *list, const char *sender)
{
	GSList *l;

	for (l = list; l; l = l->next) {
		struct session_req *req = l->data;

		if (g_str_equal(req->owner, sender))
			return req;
	}

	return NULL;
}

static uint8_t get_needed_mode(struct btd_adapter *adapter, uint8_t mode)
{
	GSList *l;

	if (adapter->global_mode > mode)
		mode = adapter->global_mode;

	for (l = adapter->mode_sessions; l; l = l->next) {
		struct session_req *req = l->data;

		if (req->mode > mode)
			mode = req->mode;
	}

	return mode;
}

static void session_remove(struct session_req *req)
{
	struct btd_adapter *adapter = req->adapter;

	if (req->mode) {
		uint8_t mode;

		adapter->mode_sessions = g_slist_remove(adapter->mode_sessions,
							req);

		mode = get_needed_mode(adapter, adapter->global_mode);

		if (mode == adapter->mode)
			return;

		debug("Switching to '%s' mode", mode2str(mode));

		set_mode(adapter, mode);
	} else {
		adapter->disc_sessions = g_slist_remove(adapter->disc_sessions,
							req);

		if (adapter->disc_sessions)
			return;

		debug("Stopping discovery");

		pending_remote_name_cancel(adapter);

		clear_found_devices_list(adapter);

		g_slist_free(adapter->oor_devices);
		adapter->oor_devices = NULL;

		if (adapter->scheduler_id)
			g_source_remove(adapter->scheduler_id);

		adapter_ops->stop_discovery(adapter->dev_id);
	}
}

static void session_free(struct session_req *req)
{
	debug("%s session %p with %s deactivated",
		req->mode ? "Mode" : "Discovery", req, req->owner);

	if (req->id)
		g_dbus_remove_watch(req->conn, req->id);

	session_remove(req);

	if (req->msg)
		dbus_message_unref(req->msg);
	if (req->conn)
		dbus_connection_unref(req->conn);
	g_free(req->owner);
	g_free(req);
}

static void session_owner_exit(DBusConnection *conn, void *user_data)
{
	struct session_req *req = user_data;

	req->id = 0;

	session_free(req);
}

static struct session_req *session_ref(struct session_req *req)
{
	req->refcount++;

	debug("session_ref(%p): ref=%d", req, req->refcount);

	return req;
}

static void session_unref(struct session_req *req)
{
	req->refcount--;

	debug("session_unref(%p): ref=%d", req, req->refcount);

	if (req->refcount)
		return;

	session_free(req);
}

static struct session_req *create_session(struct btd_adapter *adapter,
					DBusConnection *conn, DBusMessage *msg,
					uint8_t mode, GDBusWatchFunction cb)
{
	struct session_req *req;
	const char *sender = dbus_message_get_sender(msg);

	req = g_new0(struct session_req, 1);
	req->adapter = adapter;
	req->conn = dbus_connection_ref(conn);
	req->msg = dbus_message_ref(msg);
	req->owner = g_strdup(dbus_message_get_sender(msg));
	req->mode = mode;

	if (cb)
		req->id = g_dbus_add_disconnect_watch(conn, sender, cb, req,
							NULL);

	info("%s session %p with %s activated",
		req->mode ? "Mode" : "Discovery", req, sender);

	return session_ref(req);
}

static void confirm_mode_cb(struct agent *agent, DBusError *derr, void *data)
{
	struct session_req *req = data;
	int err;
	DBusMessage *reply;

	if (derr && dbus_error_is_set(derr)) {
		reply = dbus_message_new_error(req->msg, derr->name,
						derr->message);
		g_dbus_send_message(req->conn, reply);
		session_unref(req);
		return;
	}

	err = set_mode(req->adapter, req->mode);
	if (err < 0)
		reply = failed_strerror(req->msg, -err);
	else
		reply = dbus_message_new_method_return(req->msg);

	g_dbus_send_message(req->conn, reply);

	dbus_message_unref(req->msg);
	req->msg = NULL;

	if (!find_session(req->adapter->mode_sessions, req->owner))
		session_unref(req);
}

static DBusMessage *set_discoverable_timeout(DBusConnection *conn,
							DBusMessage *msg,
							uint32_t timeout,
							void *data)
{
	struct btd_adapter *adapter = data;
	const char *path;

	if (adapter->discov_timeout == timeout && timeout == 0)
		return dbus_message_new_method_return(msg);

	if (adapter->scan_mode & SCAN_INQUIRY)
		adapter_set_discov_timeout(adapter, timeout);

	adapter->discov_timeout = timeout;

	write_discoverable_timeout(&adapter->bdaddr, timeout);

	path = dbus_message_get_path(msg);

	emit_property_changed(conn, path,
				ADAPTER_INTERFACE, "DiscoverableTimeout",
				DBUS_TYPE_UINT32, &timeout);

	return dbus_message_new_method_return(msg);
}

static DBusMessage *set_pairable_timeout(DBusConnection *conn,
						DBusMessage *msg,
						uint32_t timeout,
						void *data)
{
	struct btd_adapter *adapter = data;
	const char *path;

	if (adapter->pairable_timeout == timeout && timeout == 0)
		return dbus_message_new_method_return(msg);

	if (adapter->pairable)
		adapter_set_pairable_timeout(adapter, timeout);

	adapter->pairable_timeout = timeout;

	write_pairable_timeout(&adapter->bdaddr, timeout);

	path = dbus_message_get_path(msg);

	emit_property_changed(conn, path,
				ADAPTER_INTERFACE, "PairableTimeout",
				DBUS_TYPE_UINT32, &timeout);

	return dbus_message_new_method_return(msg);
}

static void update_ext_inquiry_response(struct btd_adapter *adapter)
{
	uint8_t fec = 0, data[240];
	struct hci_dev *dev = &adapter->dev;
	int dd;

	if (!(dev->features[6] & LMP_EXT_INQ))
		return;

	memset(data, 0, sizeof(data));

	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return;

	if (dev->ssp_mode > 0)
		create_ext_inquiry_response((char *) dev->name, data);

	if (hci_write_ext_inquiry_response(dd, fec, data,
						HCI_REQ_TIMEOUT) < 0)
		error("Can't write extended inquiry response: %s (%d)",
						strerror(errno), errno);

	hci_close_dev(dd);
}

void adapter_update_local_name(bdaddr_t *bdaddr, uint8_t status, void *ptr)
{
	read_local_name_rp rp;
	struct hci_dev *dev;
	struct btd_adapter *adapter;
	gchar *name;

	if (status)
		return;

	adapter = manager_find_adapter(bdaddr);
	if (!adapter) {
		error("Unable to find matching adapter");
		return;
	}

	dev = &adapter->dev;

	memcpy(&rp, ptr, MAX_NAME_LENGTH);
	if (strncmp((char *) rp.name, (char *) dev->name, MAX_NAME_LENGTH) == 0)
		return;

	strncpy((char *) dev->name, (char *) rp.name, MAX_NAME_LENGTH);

	write_local_name(bdaddr, (char *) dev->name);

	update_ext_inquiry_response(adapter);

	name = g_strdup((char *) dev->name);

	if (connection)
		emit_property_changed(connection, adapter->path, ADAPTER_INTERFACE,
				"Name", DBUS_TYPE_STRING, &name);
	g_free(name);
}

void adapter_setname_complete(bdaddr_t *local, uint8_t status)
{
	struct btd_adapter *adapter;
	int err;

	if (status)
		return;

	adapter = manager_find_adapter(local);
	if (!adapter) {
		error("No matching adapter found");
		return;
	}

	err = adapter_ops->read_name(adapter->dev_id);
	if (err < 0)
		error("Sending getting name command failed: %s (%d)",
						strerror(errno), errno);

}

static DBusMessage *set_name(DBusConnection *conn, DBusMessage *msg,
					const char *name, void *data)
{
	struct btd_adapter *adapter = data;
	struct hci_dev *dev = &adapter->dev;
	int err;

	if (!g_utf8_validate(name, -1, NULL)) {
		error("Name change failed: supplied name isn't valid UTF-8");
		return invalid_args(msg);
	}

	if (strncmp(name, (char *) dev->name, MAX_NAME_LENGTH) == 0)
		goto done;

	if (!adapter->up)
		return failed_strerror(msg, -EHOSTDOWN);

	err = adapter_ops->set_name(adapter->dev_id, name);
	if (err < 0)
		return failed_strerror(msg, err);

done:
	return dbus_message_new_method_return(msg);
}

struct btd_device *adapter_find_device(struct btd_adapter *adapter,
							const char *dest)
{
	struct btd_device *device;
	GSList *l;

	if (!adapter)
		return NULL;

	l = g_slist_find_custom(adapter->devices, dest,
					(GCompareFunc) device_address_cmp);
	if (!l)
		return NULL;

	device = l->data;

	return device;
}

struct btd_device *adapter_find_connection(struct btd_adapter *adapter,
						uint16_t handle)
{
	GSList *l;

	for (l = adapter->connections; l; l = l->next) {
		struct btd_device *device = l->data;

		if (device_has_connection(device, handle))
			return device;
	}

	return NULL;
}

static void adapter_update_devices(struct btd_adapter *adapter)
{
	char **devices;
	int i;
	GSList *l;

	/* Devices */
	devices = g_new0(char *, g_slist_length(adapter->devices) + 1);
	for (i = 0, l = adapter->devices; l; l = l->next, i++) {
		struct btd_device *dev = l->data;
		devices[i] = (char *) device_get_path(dev);
	}

	emit_array_property_changed(connection, adapter->path,
					ADAPTER_INTERFACE, "Devices",
					DBUS_TYPE_OBJECT_PATH, &devices);
	g_free(devices);
}

struct btd_device *adapter_create_device(DBusConnection *conn,
						struct btd_adapter *adapter,
						const char *address)
{
	struct btd_device *device;
	const char *path;

	debug("adapter_create_device(%s)", address);

	device = device_create(conn, adapter, address);
	if (!device)
		return NULL;

	device_set_temporary(device, TRUE);

	adapter->devices = g_slist_append(adapter->devices, device);

	path = device_get_path(device);
	g_dbus_emit_signal(conn, adapter->path,
			ADAPTER_INTERFACE, "DeviceCreated",
			DBUS_TYPE_OBJECT_PATH, &path,
			DBUS_TYPE_INVALID);

	adapter_update_devices(adapter);

	return device;
}

void adapter_remove_device(DBusConnection *conn, struct btd_adapter *adapter,
				struct btd_device *device)
{
	const gchar *dev_path = device_get_path(device);
	struct agent *agent;

	adapter->devices = g_slist_remove(adapter->devices, device);
	adapter->connections = g_slist_remove(adapter->connections, device);

	adapter_update_devices(adapter);

	g_dbus_emit_signal(conn, adapter->path,
			ADAPTER_INTERFACE, "DeviceRemoved",
			DBUS_TYPE_OBJECT_PATH, &dev_path,
			DBUS_TYPE_INVALID);

	agent = device_get_agent(device);
	if (!agent)
		agent = adapter->agent;

	if (agent && device_is_authorizing(device))
		agent_cancel(agent);

	agent = device_get_agent(device);

	if (agent) {
		agent_destroy(agent, FALSE);
		device_set_agent(device, NULL);
	}

	device_remove(device, conn, TRUE);
}

struct btd_device *adapter_get_device(DBusConnection *conn,
						struct btd_adapter *adapter,
						const gchar *address)
{
	struct btd_device *device;

	debug("adapter_get_device(%s)", address);

	if (!adapter)
		return NULL;

	device = adapter_find_device(adapter, address);
	if (device)
		return device;

	return adapter_create_device(conn, adapter, address);
}

static int adapter_start_inquiry(struct btd_adapter *adapter)
{
	gboolean periodic = TRUE;

	if (main_opts.discov_interval)
		periodic = FALSE;

	pending_remote_name_cancel(adapter);

	return adapter_ops->start_discovery(adapter->dev_id, periodic);
}

static DBusMessage *adapter_start_discovery(DBusConnection *conn,
						DBusMessage *msg, void *data)
{
	struct session_req *req;
	struct btd_adapter *adapter = data;
	const char *sender = dbus_message_get_sender(msg);
	int err;

	if (!adapter->up)
		return adapter_not_ready(msg);

	req = find_session(adapter->disc_sessions, sender);
	if (req) {
		session_ref(req);
		return dbus_message_new_method_return(msg);
	}

	if (adapter->disc_sessions)
		goto done;

	if (main_opts.name_resolv)
		adapter->state |= RESOLVE_NAME;

	err = adapter_start_inquiry(adapter);
	if (err < 0)
		return failed_strerror(msg, -err);

done:
	req = create_session(adapter, conn, msg, 0,
				session_owner_exit);

	adapter->disc_sessions = g_slist_append(adapter->disc_sessions, req);

	return dbus_message_new_method_return(msg);
}

static DBusMessage *adapter_stop_discovery(DBusConnection *conn,
						DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;
	struct session_req *req;
	const char *sender = dbus_message_get_sender(msg);

	if (!adapter->up)
		return adapter_not_ready(msg);

	req = find_session(adapter->disc_sessions, sender);
	if (!req)
		return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
				"Invalid discovery session");

	session_unref(req);
	info("Stopping discovery");
	return dbus_message_new_method_return(msg);
}

struct remote_device_list_t {
	GSList *list;
	time_t time;
};

static DBusMessage *get_properties(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;
	const char *property;
	DBusMessage *reply;
	DBusMessageIter iter;
	DBusMessageIter dict;
	char str[MAX_NAME_LENGTH + 1], srcaddr[18];
	uint32_t class;
	gboolean value;
	char **devices;
	int i;
	GSList *l;

	ba2str(&adapter->bdaddr, srcaddr);

	if (check_address(srcaddr) < 0)
		return adapter_not_ready(msg);

	reply = dbus_message_new_method_return(msg);
	if (!reply)
		return NULL;

	dbus_message_iter_init_append(reply, &iter);

	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
			DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);

	/* Address */
	property = srcaddr;
	dict_append_entry(&dict, "Address", DBUS_TYPE_STRING, &property);

	/* Name */
	memset(str, 0, sizeof(str));
	strncpy(str, (char *) adapter->dev.name, MAX_NAME_LENGTH);
	property = str;

	dict_append_entry(&dict, "Name", DBUS_TYPE_STRING, &property);

	/* Class */
	class = adapter->dev.class[0] |
			adapter->dev.class[1] << 8 |
			adapter->dev.class[2] << 16;
	dict_append_entry(&dict, "Class", DBUS_TYPE_UINT32, &class);

	/* Powered */
	value = (adapter->up && !adapter->off_requested) ? TRUE : FALSE;
	dict_append_entry(&dict, "Powered", DBUS_TYPE_BOOLEAN, &value);

	/* Discoverable */
	value = adapter->scan_mode & SCAN_INQUIRY ? TRUE : FALSE;
	dict_append_entry(&dict, "Discoverable", DBUS_TYPE_BOOLEAN, &value);

	/* Pairable */
	dict_append_entry(&dict, "Pairable", DBUS_TYPE_BOOLEAN,
				&adapter->pairable);

	/* DiscoverableTimeout */
	dict_append_entry(&dict, "DiscoverableTimeout",
				DBUS_TYPE_UINT32, &adapter->discov_timeout);

	/* PairableTimeout */
	dict_append_entry(&dict, "PairableTimeout",
				DBUS_TYPE_UINT32, &adapter->pairable_timeout);


	if (adapter->state & PERIODIC_INQUIRY || adapter->state & STD_INQUIRY)
		value = TRUE;
	else
		value = FALSE;

	/* Discovering */
	dict_append_entry(&dict, "Discovering", DBUS_TYPE_BOOLEAN, &value);

	/* Devices */
	devices = g_new0(char *, g_slist_length(adapter->devices) + 1);
	for (i = 0, l = adapter->devices; l; l = l->next, i++) {
		struct btd_device *dev = l->data;
		devices[i] = (char *) device_get_path(dev);
	}
	dict_append_array(&dict, "Devices", DBUS_TYPE_OBJECT_PATH,
								&devices, i);
	g_free(devices);

	dbus_message_iter_close_container(&iter, &dict);

	return reply;
}

static DBusMessage *set_property(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;
	DBusMessageIter iter;
	DBusMessageIter sub;
	const char *property;
	char srcaddr[18];

	ba2str(&adapter->bdaddr, srcaddr);

	if (!dbus_message_iter_init(msg, &iter))
		return invalid_args(msg);

	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_STRING)
		return invalid_args(msg);

	dbus_message_iter_get_basic(&iter, &property);
	dbus_message_iter_next(&iter);

	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_VARIANT)
		return invalid_args(msg);
	dbus_message_iter_recurse(&iter, &sub);

	if (g_str_equal("Name", property)) {
		const char *name;

		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_STRING)
			return invalid_args(msg);
		dbus_message_iter_get_basic(&sub, &name);

		return set_name(conn, msg, name, data);
	} else if (g_str_equal("Powered", property)) {
		gboolean powered;

		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_BOOLEAN)
			return invalid_args(msg);

		dbus_message_iter_get_basic(&sub, &powered);

		return set_powered(conn, msg, powered, data);
	} else if (g_str_equal("Discoverable", property)) {
		gboolean discoverable;

		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_BOOLEAN)
			return invalid_args(msg);

		dbus_message_iter_get_basic(&sub, &discoverable);

		return set_discoverable(conn, msg, discoverable, data);
	} else if (g_str_equal("DiscoverableTimeout", property)) {
		uint32_t timeout;

		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_UINT32)
			return invalid_args(msg);

		dbus_message_iter_get_basic(&sub, &timeout);

		return set_discoverable_timeout(conn, msg, timeout, data);
	} else if (g_str_equal("Pairable", property)) {
		gboolean pairable;

		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_BOOLEAN)
			return invalid_args(msg);

		dbus_message_iter_get_basic(&sub, &pairable);

		return set_pairable(conn, msg, pairable, data);
	} else if (g_str_equal("PairableTimeout", property)) {
		uint32_t timeout;

		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_UINT32)
			return invalid_args(msg);

		dbus_message_iter_get_basic(&sub, &timeout);

		return set_pairable_timeout(conn, msg, timeout, data);
	}

	return invalid_args(msg);
}

static DBusMessage *request_session(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;
	struct session_req *req;
	const char *sender = dbus_message_get_sender(msg);
	uint8_t new_mode;
	int ret;

	if (!adapter->agent)
		return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
						"No agent registered");

	if (!adapter->mode_sessions)
		adapter->global_mode = adapter->mode;

	new_mode = get_mode(&adapter->bdaddr, "on");

	req = find_session(adapter->mode_sessions, sender);
	if (req) {
		session_ref(req);
		return dbus_message_new_method_return(msg);
	} else {
		req = create_session(adapter, conn, msg, new_mode,
					session_owner_exit);
		adapter->mode_sessions = g_slist_append(adapter->mode_sessions,
							req);
	}

	/* No need to change mode */
	if (adapter->mode >= new_mode)
		return dbus_message_new_method_return(msg);

	ret = agent_confirm_mode_change(adapter->agent, mode2str(new_mode),
					confirm_mode_cb, req, NULL);
	if (ret < 0) {
		session_unref(req);
		return failed_strerror(msg, -ret);
	}

	return NULL;
}

static DBusMessage *release_session(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;
	struct session_req *req;
	const char *sender = dbus_message_get_sender(msg);

	req = find_session(adapter->mode_sessions, sender);
	if (!req)
		return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
				"No Mode to release");

	session_unref(req);

	return dbus_message_new_method_return(msg);
}

static DBusMessage *list_devices(DBusConnection *conn,
						DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;
	DBusMessage *reply;
	GSList *l;
	DBusMessageIter iter;
	DBusMessageIter array_iter;
	const gchar *dev_path;

	if (!dbus_message_has_signature(msg, DBUS_TYPE_INVALID_AS_STRING))
		return invalid_args(msg);

	reply = dbus_message_new_method_return(msg);
	if (!reply)
		return NULL;

	dbus_message_iter_init_append(reply, &iter);
	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
				DBUS_TYPE_OBJECT_PATH_AS_STRING, &array_iter);

	for (l = adapter->devices; l; l = l->next) {
		struct btd_device *device = l->data;

		dev_path = device_get_path(device);

		dbus_message_iter_append_basic(&array_iter,
				DBUS_TYPE_OBJECT_PATH, &dev_path);
	}

	dbus_message_iter_close_container(&iter, &array_iter);

	return reply;
}

static DBusMessage *cancel_device_creation(DBusConnection *conn,
						DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;
	const gchar *address, *sender = dbus_message_get_sender(msg);
	struct btd_device *device;

	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &address,
						DBUS_TYPE_INVALID) == FALSE)
		return invalid_args(msg);

	if (check_address(address) < 0)
		return invalid_args(msg);

	device = adapter_find_device(adapter, address);
	if (!device || !device_is_creating(device, NULL))
		return g_dbus_create_error(msg,
				ERROR_INTERFACE ".NotInProgress",
				"Device creation not in progress");

	if (!device_is_creating(device, sender))
		return not_authorized(msg);

	device_set_temporary(device, TRUE);

	if (device_is_connected(device)) {
		device_request_disconnect(device, msg);
		return NULL;
	}

	adapter_remove_device(conn, adapter, device);

	return dbus_message_new_method_return(msg);
}

static DBusMessage *create_device(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;
	struct btd_device *device;
	const gchar *address;

	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &address,
						DBUS_TYPE_INVALID) == FALSE)
		return invalid_args(msg);

	if (check_address(address) < 0)
		return invalid_args(msg);

	if (adapter_find_device(adapter, address))
		return g_dbus_create_error(msg,
				ERROR_INTERFACE ".AlreadyExists",
				"Device already exists");

	debug("create_device(%s)", address);

	device = adapter_create_device(conn, adapter, address);
	if (!device)
		return NULL;

	device_browse(device, conn, msg, NULL, FALSE);

	return NULL;
}

static uint8_t parse_io_capability(const char *capability)
{
	if (g_str_equal(capability, ""))
		return IO_CAPABILITY_DISPLAYYESNO;
	if (g_str_equal(capability, "DisplayOnly"))
		return IO_CAPABILITY_DISPLAYONLY;
	if (g_str_equal(capability, "DisplayYesNo"))
		return IO_CAPABILITY_DISPLAYYESNO;
	if (g_str_equal(capability, "KeyboardOnly"))
		return IO_CAPABILITY_KEYBOARDONLY;
	if (g_str_equal(capability, "NoInputNoOutput"))
		return IO_CAPABILITY_NOINPUTNOOUTPUT;
	return IO_CAPABILITY_INVALID;
}

static DBusMessage *create_paired_device(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;
	struct btd_device *device;
	const gchar *address, *agent_path, *capability, *sender;
	uint8_t cap;

	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &address,
					DBUS_TYPE_OBJECT_PATH, &agent_path,
					DBUS_TYPE_STRING, &capability,
						DBUS_TYPE_INVALID) == FALSE)
		return invalid_args(msg);

	if (check_address(address) < 0)
		return invalid_args(msg);

	sender = dbus_message_get_sender(msg);
	if (adapter->agent &&
			agent_matches(adapter->agent, sender, agent_path)) {
		error("Refusing adapter agent usage as device specific one");
		return invalid_args(msg);
	}

	cap = parse_io_capability(capability);
	if (cap == IO_CAPABILITY_INVALID)
		return invalid_args(msg);

	device = adapter_get_device(conn, adapter, address);
	if (!device)
		return g_dbus_create_error(msg,
				ERROR_INTERFACE ".Failed",
				"Unable to create a new device object");

	return device_create_bonding(device, conn, msg, agent_path, cap);
}

static gint device_path_cmp(struct btd_device *device, const gchar *path)
{
	const gchar *dev_path = device_get_path(device);

	return strcasecmp(dev_path, path);
}

static DBusMessage *remove_device(DBusConnection *conn,
						DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;
	struct btd_device *device;
	const char *path;
	GSList *l;

	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_OBJECT_PATH, &path,
						DBUS_TYPE_INVALID) == FALSE)
		return invalid_args(msg);

	l = g_slist_find_custom(adapter->devices,
			path, (GCompareFunc) device_path_cmp);
	if (!l)
		return g_dbus_create_error(msg,
				ERROR_INTERFACE ".DoesNotExist",
				"Device does not exist");
	device = l->data;

	if (device_is_temporary(device) || device_is_busy(device))
		return g_dbus_create_error(msg,
				ERROR_INTERFACE ".DoesNotExist",
				"Device creation in progress");

	device_set_temporary(device, TRUE);

	if (!device_is_connected(device)) {
		adapter_remove_device(conn, adapter, device);
		return dbus_message_new_method_return(msg);
	}

	device_request_disconnect(device, msg);
	return NULL;
}

static DBusMessage *find_device(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;
	struct btd_device *device;
	DBusMessage *reply;
	const gchar *address;
	GSList *l;
	const gchar *dev_path;

	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &address,
						DBUS_TYPE_INVALID))
		return invalid_args(msg);

	l = g_slist_find_custom(adapter->devices,
			address, (GCompareFunc) device_address_cmp);
	if (!l)
		return g_dbus_create_error(msg,
				ERROR_INTERFACE ".DoesNotExist",
				"Device does not exist");

	device = l->data;

	reply = dbus_message_new_method_return(msg);
	if (!reply)
		return NULL;

	dev_path = device_get_path(device);

	dbus_message_append_args(reply,
				DBUS_TYPE_OBJECT_PATH, &dev_path,
				DBUS_TYPE_INVALID);

	return reply;
}

static void agent_removed(struct agent *agent, struct btd_adapter *adapter)
{
	adapter->agent = NULL;
}

static DBusMessage *register_agent(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	const char *path, *name, *capability;
	struct agent *agent;
	struct btd_adapter *adapter = data;
	uint8_t cap;

	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_OBJECT_PATH, &path,
			DBUS_TYPE_STRING, &capability, DBUS_TYPE_INVALID))
		return NULL;

	if (adapter->agent)
		return g_dbus_create_error(msg,
				ERROR_INTERFACE ".AlreadyExists",
				"Agent already exists");

	cap = parse_io_capability(capability);
	if (cap == IO_CAPABILITY_INVALID)
		return invalid_args(msg);

	name = dbus_message_get_sender(msg);

	agent = agent_create(adapter, name, path, cap,
				(agent_remove_cb) agent_removed, adapter);
	if (!agent)
		return g_dbus_create_error(msg,
				ERROR_INTERFACE ".Failed",
				"Failed to create a new agent");

	adapter->agent = agent;

	debug("Agent registered for hci%d at %s:%s", adapter->dev_id, name,
			path);

	return dbus_message_new_method_return(msg);
}

static DBusMessage *unregister_agent(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	const char *path, *name;
	struct btd_adapter *adapter = data;

	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_OBJECT_PATH, &path,
						DBUS_TYPE_INVALID))
		return NULL;

	name = dbus_message_get_sender(msg);

	if (!adapter->agent || !agent_matches(adapter->agent, name, path))
		return g_dbus_create_error(msg,
				ERROR_INTERFACE ".DoesNotExist",
				"No such agent");

	agent_destroy(adapter->agent, FALSE);
	adapter->agent = NULL;

	return dbus_message_new_method_return(msg);
}

static sdp_record_t *create_rfcomm_record(struct btd_adapter *adapter,
					const char *name, uuid_t uuid, uint8_t channel)
{
	uuid_t root_uuid, l2cap_uuid, rfcomm_uuid;
	sdp_list_t *svclass, *root, *proto;
	sdp_record_t *record;

	record = sdp_record_alloc();
	if (!record)
		return NULL;

	sdp_uuid16_create(&root_uuid, PUBLIC_BROWSE_GROUP);
	root = sdp_list_append(NULL, &root_uuid);
	sdp_set_browse_groups(record, root);

	sdp_uuid16_create(&l2cap_uuid, L2CAP_UUID);
	proto = sdp_list_append(NULL, sdp_list_append(NULL, &l2cap_uuid));

	sdp_uuid16_create(&rfcomm_uuid, RFCOMM_UUID);
	proto = sdp_list_append(proto, sdp_list_append(
			sdp_list_append(NULL, &rfcomm_uuid),
			sdp_data_alloc(SDP_UINT8, &channel)));

	sdp_set_access_protos(record, sdp_list_append(NULL, proto));

	svclass = sdp_list_append(NULL, &uuid);
	sdp_set_service_classes(record, svclass);

	sdp_set_info_attr(record, name, NULL, NULL);

	return record;
}

static DBusMessage *add_rfcomm_service_record(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	uuid_t uuid;
	const char *name;
	uint8_t channel;
	uint32_t *uuid_p;
	uint32_t uuid_net[4];   // network order
	uint64_t uuid_host[2];  // host
	sdp_record_t *record;
	struct btd_adapter *adapter = data;

	DBusMessage *reply;

	if (!dbus_message_get_args(msg, NULL,
			DBUS_TYPE_STRING, &name,
			DBUS_TYPE_UINT64, &uuid_host[0],
			DBUS_TYPE_UINT64, &uuid_host[1],
			DBUS_TYPE_UINT16, &channel,
			DBUS_TYPE_INVALID))
		return invalid_args(msg);

	uuid_p = (uint32_t *)uuid_host;
	uuid_net[1] = htonl(*uuid_p++);
	uuid_net[0] = htonl(*uuid_p++);
	uuid_net[3] = htonl(*uuid_p++);
	uuid_net[2] = htonl(*uuid_p++);

	sdp_uuid128_create(&uuid, (void *)uuid_net);

	record = create_rfcomm_record(adapter, name, uuid, channel);

	if (!record)
		return g_dbus_create_error(msg,
				ERROR_INTERFACE ".Failed",
				"Failed to create sdp record");

	if (add_record_to_server(&adapter->bdaddr, record))
		return g_dbus_create_error(msg,
				ERROR_INTERFACE ".Failed",
				"Failed to register sdp record");

	reply = dbus_message_new_method_return(msg);
	dbus_message_append_args(reply,
			DBUS_TYPE_UINT32, &record->handle,
			DBUS_TYPE_INVALID);

	return reply;
}

static DBusMessage *remove_service_record(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;
	dbus_uint32_t handle;

	if (!dbus_message_get_args(msg, NULL,
			DBUS_TYPE_UINT32, &handle,
			DBUS_TYPE_INVALID))
		return invalid_args(msg);

	if (remove_record_from_server(handle))
		return g_dbus_create_error(msg,
				ERROR_INTERFACE ".Failed",
				"Failed to remove sdp record");

	return dbus_message_new_method_return(msg);
}
//#ifdef CONFIG_GEMINI_GENERIC


//20090924 RDS function
static void BCM4325_updateRDSInfo(uint8_t *data)
{
    uint8_t data_index,rds_index;
	uint8_t data_length = BCM4325_RDS_FIFO_WLINE*BCM4325_RDS_FIFO_DUPLE_SIZE;
	BCM4325_RDS_PARAMETERS sFM_RDSData[BCM4325_RDS_FIFO_WLINE];

    for (rds_index=0,data_index = 0; data_index< data_length; data_index++)
    {
		if ( data_index%3 == 0)
		{		 		   
            //bites[7:4] RDS block type     
            //bites[3:2] RDS block quality descriptor
            sFM_RDSData[rds_index].rds_block_id = (data[data_index]&BCM4325_RDS_BLOCK_MASK);
            sFM_RDSData[rds_index].rds_block_errors = ((data[data_index]&BCM4325_RDS_QUALITY_MASK)>>2);          
			LOGE("data[%d] rds_block_id=[0x%02x] rds_block_errors=[0x%02x] ",data_index,sFM_RDSData[rds_index].rds_block_id,sFM_RDSData[rds_index].rds_block_errors);
		}
		else if ( data_index % 3 == 1 )//information
		{
            sFM_RDSData[rds_index].rds_data_block = ((data[data_index]&0xff) << 8);
		}
		else
		{
	        sFM_RDSData[rds_index].rds_data_block |= (data[data_index]&0xff);  
			LOGE("                                rds_data_block=[0x%02x] ",sFM_RDSData[rds_index].rds_data_block);
    	    rds_index += 1;
		}
    }

    RDSDISegmentAddr = 0xff;
    RDSPrevBlockID = BCM4325_RDS_BLOCK_INVALID;
    RDSNextBlockID = sFM_RDSData[1].rds_block_id;
    BCM4325_analysisRDSData(&sFM_RDSData[0]);
    for (rds_index = 1; rds_index < (BCM4325_RDS_FIFO_WLINE - 1); rds_index++)
    {
        RDSPrevBlockID = sFM_RDSData[rds_index-1].rds_block_id;
        RDSNextBlockID = sFM_RDSData[rds_index+1].rds_block_id;
        BCM4325_analysisRDSData(&sFM_RDSData[rds_index]);
    }

}


static int BCM4325_analysisRDSData(BCM4325_RDS_PARAMETERS *pRDS_parameters)
{
    BCM4325_RDS_PARAMETERS local_rds_data = *pRDS_parameters;
	int re = -1;

    //if (local_rds_data.rds_block_errors == BCM4325_RDS_DE_UNCORRECTABLE)
    if (local_rds_data.rds_block_errors != BCM4325_RDS_DE_0_ERR)
	{
        //LOGE("rds_block_id=%#x  UNCORRECTABLE ",local_rds_data.rds_block_id);
        LOGI("rds_block_id=%#x  rds data error ",local_rds_data.rds_block_id);
		if (local_rds_data.rds_block_id == BCM4325_RDS_BLOCK_B)
			   RDSDISegmentAddr = 0xff;
		return re;
    }
    switch (local_rds_data.rds_block_id)
    {
        case BCM4325_RDS_BLOCK_A:
            if (RDSNextBlockID != BCM4325_RDS_BLOCK_B)
            {
                 RDSInvalidBlockB = 1;
            }
            else if ((RDSPrevBlockID == BCM4325_RDS_BLOCK_D) && (RDSNextBlockID == BCM4325_RDS_BLOCK_B))
            {
                RDSPrevPI = RDSPICode;
                RDSCurrentPI = local_rds_data.rds_data_block;
                
                //RETAILMSG(1, (L"Current PI: %04x \r\n", s_RDSCurrentPI));
                if (RDSPrevPI == RDSCurrentPI)
                {
                    RDSPIMatched = 1;
                }
                else
                {
                    LOGE("RDS PI code un-matched");
                    RDSPIMatched = 0;
                    RDSPICode = local_rds_data.rds_data_block;                    
                }
            }
			re=1;
            break;

        case BCM4325_RDS_BLOCK_B:
            if ((RDSPrevBlockID == BCM4325_RDS_BLOCK_A) 
                && ((RDSNextBlockID == BCM4325_RDS_BLOCK_C)
                ||(RDSNextBlockID == BCM4325_RDS_BLOCK_C_SCORED)))
            {
                RDSInvalidBlockB = 0;
                RDSGroupCode = ((local_rds_data.rds_data_block & 0xF800)>>11);
                if ((RDSGroupCode == BCM4325_RDS_GROUP_A)||(RDSGroupCode == BCM4325_RDS_GROUP_B	))
			   {
                    RDSDISegmentAddr = (local_rds_data.rds_data_block) & 0x03;
					LOGI(" Block B ==>RDSDISegmentAddr=%d ",RDSDISegmentAddr);
                }
				else
			    {
			        LOGI(" Block B ==>not group 0A or 0B");
					RDSDISegmentAddr = 0xff;
                }
            }
			else
            {
                   LOGI(" Block B ==>RDSInvalidBlockB");
				   //RDSDISegmentAddr = 0xff;
				   RDSInvalidBlockB = 1;
			}
			re=1;
            break;

        case BCM4325_RDS_BLOCK_D:
            if (((RDSPrevBlockID == BCM4325_RDS_BLOCK_C)
                ||(RDSPrevBlockID == BCM4325_RDS_BLOCK_C_SCORED))
                && (RDSNextBlockID == BCM4325_RDS_BLOCK_A))
            {
                //LOGI("PrevBlockID and NextBlockID check pass!!!");
                if ((RDSInvalidBlockB == 0)&& (RDSPIMatched == 1) )
                {
					LOGI(" Block D ==>RDSDISegmentAddr=%d data=[%#x]",RDSDISegmentAddr,local_rds_data.rds_data_block);
                    BCM4325_UpdatePSName(RDSDISegmentAddr, local_rds_data.rds_data_block);
                }
            }
			re=1;
            break;

        default:
            return re;
    }

    return re;
}

/* 
*   BCM4325_UpdatePSName : Update Program Service Name
*/
static int BCM4325_UpdatePSName( 
    uint8_t DIAddr,
    uint16_t PSData
)
{
    uint8_t low_byte, high_byte;
    uint16_t dwData;

    dwData = PSData;

    //RETAILMSG(TRUE, (L"data: %x \r\n", dwData));
    if ((RDSGroupCode == BCM4325_RDS_GROUP_A) || (RDSGroupCode==BCM4325_RDS_GROUP_B))
	{
        high_byte = msb(PSData);
        low_byte =  lsb(PSData);
        
        //RETAILMSG(TRUE, (L"HI: %x LO %x\r\n", high_byte, low_byte));
        
        if ((low_byte<0x20)||(low_byte>0x9F)) low_byte=0x20;    /* INVALID so Assign to NULL */
        if ((high_byte<0x20)||(high_byte>0x9F)) high_byte=0x20;    /* INVALID so Assign to NULL */
        
        switch(DIAddr)
        {
            case 0:
                RDSProgramServiceName[0] = high_byte;
                RDSProgramServiceName[1] = low_byte;
                break;
            case 1:
                RDSProgramServiceName[2] = high_byte;
                RDSProgramServiceName[3] = low_byte;
                break;
            case 2:
                RDSProgramServiceName[4] = high_byte;
                RDSProgramServiceName[5] = low_byte;
                break;
            case 3:
                RDSProgramServiceName[6] = high_byte;
                RDSProgramServiceName[7] = low_byte;
                break;
        }
    }
    return 1;
}

static void BCM4325_getPSName( uint8_t *pRDSData)
{
/*
    RETAILMSG(TRUE, (L"GetRDS %c%c%c%c%c%c%c%c \r\n",
         RDSProgramServiceName[0],
         RDSProgramServiceName[1],
         RDSProgramServiceName[2],
         RDSProgramServiceName[3],
         RDSProgramServiceName[4],
         RDSProgramServiceName[5],
         RDSProgramServiceName[6],
         RDSProgramServiceName[7]
        ));
*/   
    int i;
    for ( i = 0; i < BCM4325_RDS_PSN_LENGTH; i++ ){
        pRDSData[i] = RDSProgramServiceName[i];
    }

}

//RDS function end

/* bcm_vsc_write_reg
Read BCM Vendor Specific Register commnad 
example:
if (bcm_vsc_write_reg(dd, address, value, 1000) < 0) {
	perror("HCI bcm_vsc_write_reg request failed");
	exit(1);
}*/
static int bcm_vsc_write_reg(int dd, uint8_t address, uint8_t value, int to)
{
	bcm_write_vsc_cmd_cp cp;
	bcm_write_vsc_cmd_rp rp;
	struct hci_request rq;
	int len = 0x03;

    if ( address == 0xfd )
		len = 0x04;
  
	memset(&cp, 0, sizeof(cp));
	cp.addr		= address;
	cp.dir		= BCM4325_I2C_FM_REG_WRITE;	
	if (len == 0x03)
	    cp.param[0] 	= value & 0xff;
    else
    {
       cp.param[0] 	= value;  // low byte
	   cp.param[1] 	= 0x00;   // high byte
    }
	
	memset(&rq, 0, sizeof(rq));
	rq.ogf    = OGF_VENDOR_CMD;
	rq.ocf    = BCM4325_OPCODE_OCF;
	rq.cparam = &cp;
	//rq.clen   = 0x03;	//BCM4325_WRITE_VSC_HCI_CMD_CP_SIZE - 1;
    rq.clen   = len;
	rq.rparam = &rp;
	rq.rlen   = BCM4325_WRITE_VSC_HCI_CMD_RP_SIZE;
	
	debug("bcm_vsc_write_reg(Reg:0x%02X,(rw=%d) = 0x%02X)",cp.addr, cp.dir, cp.param[0]);
	
	if (hci_send_req(dd, &rq, to) < 0)
		return -1;

	if (rp.status) {
		errno = EIO;
		return -1;
	}
	
	return 0;
}


/* bcm_vsc_read_reg
Read BCM Vendor Specific Register commnad 
example:
if (bcm_vsc_read_reg(dd, address, &value, 1000) < 0) {
	perror("HCI bcm_vsc_read_reg request failed");
	exit(1);
}*/
static int bcm_vsc_read_reg(int dd, uint8_t address, uint8_t *value, int to)
{
	bcm_read_vsc_cmd_cp cp;
	bcm_read_vsc_cmd_rp rp;
	int ret;
	struct hci_request rq;

	memset(&cp, 0, sizeof(cp));
	cp.addr		= address;
	cp.dir		= BCM4325_I2C_FM_REG_READ;	
	cp.length		= 0x01;	

	memset(&rq, 0, sizeof(rq));
	rq.ogf    = OGF_VENDOR_CMD;
	rq.ocf    = BCM4325_OPCODE_OCF;
	rq.cparam = &cp;
	rq.clen   = BCM4325_READ_VSC_HCI_CMD_CP_SIZE;
	rq.rparam = &rp;
	rq.rlen   = 0x04;	//M4325_READ_VSC_HCI_CMD_RP_SIZE - 1;

	if (hci_send_req(dd, &rq, to) < 0)
		return -1;

	if (rp.status) {
		errno = EIO;
		return -1;
	}
	
	*value = rp.param[0];
		
	debug("bcm_vsc_read_reg(Reg:0x%02X,(rw=%d) = 0x%02X)",cp.addr, cp.dir, rp.param[0]);	

	return 0;
}

/* bcm_vsc_read_rds_reg
Read BCM Vendor Specific Register commnad for get rds data
example:
if (bcm_vsc_read_reg(dd, address, &value, 1000) < 0) {
	perror("HCI bcm_vsc_read_reg request failed");
	exit(1);
}*/
static int bcm_vsc_read_rds_reg(int dd, uint8_t address, uint8_t *value, int to)
{
	bcm_read_vsc_cmd_cp cp;
	bcm_read_vsc_cmd_rds_rp rp;
	int ret,i;
	struct hci_request rq;

	memset(&cp, 0, sizeof(cp));
	cp.addr		= address;
	cp.dir		= BCM4325_I2C_FM_REG_READ;	
	cp.length		= BCM4325_RDS_FIFO_WLINE*BCM4325_RDS_FIFO_DUPLE_SIZE;	

	memset(&rq, 0, sizeof(rq));
	rq.ogf    = OGF_VENDOR_CMD;
	rq.ocf    = BCM4325_OPCODE_OCF;
	rq.cparam = &cp;
	rq.clen   = BCM4325_READ_VSC_HCI_CMD_CP_SIZE;
	rq.rparam = &rp;
	rq.rlen   = BCM4325_READ_VSC_HCI_CMD_RDS_RP_SIZE;

	if (hci_send_req(dd, &rq, to) < 0)
		return -1;

	if (rp.status) {
		errno = EIO;
		return -1;
	}

	for (i=0; i < cp.length ;i++ )
	{
	     value[i] = rp.param[i];
		 //LOGE(" value[%d]=%#x  rp.param[%d]=%#x",i,value[i],i,rp.param[i]);
	}	
	
	debug("bcm_vsc_read_reg(Reg:0x%02X,(rw=%d) = 0x%02X)",cp.addr, cp.dir, rp.param[0]);	

	return 0;
}


static DBusMessage *adapter_set_fm_power(DBusConnection *conn,
				DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;	//eastern modify
	//param mode : 1 = on,   0 = off
	uint8_t	reg_addr;
	uint32_t	set_value;
	int dd;
	
	//LOGE("eastern,adapter_set_fm_power: %s ", __FUNCTION__);
	
	debug("%s:>>",__FUNCTION__);	
	
	if (!adapter->up)
		return adapter_not_ready(msg);

	if (!dbus_message_get_args(msg, NULL,
				DBUS_TYPE_UINT32, &set_value,
				DBUS_TYPE_INVALID))
		return invalid_args(msg);
	
	debug("set_value = %d", set_value);	
	
#ifndef DISABLE_STEP2
	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return no_such_adapter(msg);
	reg_addr		= BCM4325_I2C_FM_RDS_SYSTEM;
	if(0 == set_value) {
	    set_value	= BCM4325_POWER_OFF;
		if (bcm_vsc_write_reg(dd, reg_addr, (uint8_t)set_value, 1000) < 0) {
		    int err = errno;
		    error("hci_send_req() error: %s (%d)",
				         strerror(errno), errno);
		    hci_close_dev(dd);
		    return failed_strerror(msg, err);
	     }
	}
	else {
	    set_value	= BCM4325_POWER_ON;
        if (bcm_vsc_write_reg(dd, reg_addr, (uint8_t)set_value, 1000) < 0) {
		    int err = errno;
		    error("hci_send_req() error: %s (%d)",
			    	strerror(errno), errno);
		    hci_close_dev(dd);
		    return failed_strerror(msg, err);
	    }
		
		//sleep 20ms (warm-up time)
        usleep(BCM4325_DEFAULT_POWERING_DELAY*1000);
		
       //set in auto stereo/mono mode
       set_value = BCM4325_STEREO_MONO_AUTO_SELECT;
		if (bcm_vsc_write_reg(dd, BCM4325_I2C_FM_CTRL, (uint8_t)set_value, 1000) < 0) {
			int err = errno;
			error("hci_send_req() error: %s (%d)",
					strerror(errno), errno);
			hci_close_dev(dd);
			return failed_strerror(msg, err);
		}	

		//set audio rounting to DAC & mute audio

		set_value =  BCM4325_DAC_OUTPUT_LEFT | 
		             BCM4325_DAC_OUTPUT_RIGHT | BCM4325_AUDIO_ROUTE_DAC;
		if (bcm_vsc_write_reg(dd, BCM4325_I2C_FM_AUDIO_CTRL0, (uint8_t)set_value, 1000) < 0) {
			int err = errno;
			error("hci_send_req() error: %s (%d)",
					strerror(errno), errno);
			hci_close_dev(dd);
			return failed_strerror(msg, err);
		}
		
	}	
	hci_close_dev(dd);
#endif	//DISABLE_STEP2	
	return dbus_message_new_method_return(msg);
}

static DBusMessage *adapter_get_fm_power(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	const struct btd_adapter *adapter = data;	//eastern modify
	DBusMessage *reply;
	
	uint8_t	reg_addr;
	uint8_t	get_value;	
	uint32_t ret =0;
	int dd;
	
	debug("%s:>>",__FUNCTION__);	
	
	if (!dbus_message_has_signature(msg, DBUS_TYPE_INVALID_AS_STRING))
		return invalid_args(msg);	

	reply = dbus_message_new_method_return(msg);
	if (!adapter->up)
		return adapter_not_ready(msg);

#ifndef DISABLE_STEP2

	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return no_such_adapter(msg);

	reg_addr	= BCM4325_I2C_FM_RDS_SYSTEM;
	
	if (bcm_vsc_read_reg(dd, reg_addr, &get_value, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}
	hci_close_dev(dd);
#endif	//DISABLE_STEP2

	//param  : 1 = on,   0 = off
	if((get_value & BCM4325_FM_ON) == BCM4325_FM_ON)
		ret = 1;	//ON
	else	
		ret = 0;	//OFF

	debug("		ret = %d", ret);			
	
	dbus_message_append_args(reply, DBUS_TYPE_UINT32, &ret,
					DBUS_TYPE_INVALID);	

	return reply;
}

static DBusMessage *adapter_set_fm_rdspower(DBusConnection *conn,
				DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;	//eastern modify
	//param mode : 1 = on,   0 = off
	uint8_t	reg_addr;
	uint8_t	get_value;	
	uint32_t	set_value;
	int dd;
	
	debug("%s:>>",__FUNCTION__);	
	
	if (!adapter->up)
		return adapter_not_ready(msg);

	if (!dbus_message_get_args(msg, NULL,
				DBUS_TYPE_UINT32, &set_value,
				DBUS_TYPE_INVALID))
		return invalid_args(msg);
	
	debug("set_value = %d", set_value);	

#ifndef DISABLE_STEP2
	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return no_such_adapter(msg);
	reg_addr		= BCM4325_I2C_FM_RDS_SYSTEM;
    if(0 == set_value) {
		// turn off RDS power and keep FM power on
	    set_value	= BCM4325_FM_ON;
		if (bcm_vsc_write_reg(dd, reg_addr, (uint8_t)set_value, 1000) < 0) {
		    int err = errno;
		    error("hci_send_req() error: %s (%d)",
				         strerror(errno), errno);
		    hci_close_dev(dd);
		    return failed_strerror(msg, err);
	     }
	}
	else {
	    set_value	= (BCM4325_FM_ON | BCM4325_RDS_ON);
        if (bcm_vsc_write_reg(dd, reg_addr, (uint8_t)set_value, 1000) < 0) {
		    int err = errno;
		    error("hci_send_req() error: %s (%d)",
			    	strerror(errno), errno);
		    hci_close_dev(dd);
		    return failed_strerror(msg, err);
	    }
		
		//sleep 20ms (warm-up time)
        usleep(BCM4325_DEFAULT_POWERING_DELAY*1000);

	   //Need to flush FIFO??	
		
       //set RDS FIFO interrupt waterline threshold.
       //set waterline to 50 RDS tuples.
       set_value = BCM4325_RDS_FIFO_WLINE;
		if (bcm_vsc_write_reg(dd, BCM4325_I2C_RDS_WLINE, (uint8_t)set_value, 1000) < 0) {
			int err = errno;
			error("hci_send_req() error: %s (%d)",
					strerror(errno), errno);
			hci_close_dev(dd);
			return failed_strerror(msg, err);
		}	

		//clear IRQ flag	
		if ( (bcm_vsc_read_reg(dd, BCM4325_I2C_FM_RDS_FLAG0, &get_value, 1000) < 0) || 
			  (bcm_vsc_read_reg(dd, BCM4325_I2C_FM_RDS_FLAG1, &get_value, 1000) < 0) ) {
			int err = errno;
			error("hci_send_req() error: %s (%d)",
					strerror(errno), errno);
			hci_close_dev(dd);
			return failed_strerror(msg, err);
		}

		//enable water line interrupt
		set_value = BCM4325_RDS_MASK_FIFO_WLINE;
		if (bcm_vsc_write_reg(dd, BCM4325_I2C_FM_RDS_MASK1, (uint8_t)set_value, 1000) < 0) {
			int err = errno;
			error("hci_send_req() error: %s (%d)",
					strerror(errno), errno);
			hci_close_dev(dd);
			return failed_strerror(msg, err);
		}
		
	}	

	hci_close_dev(dd);
#endif	//DISABLE_STEP2	
	return dbus_message_new_method_return(msg);
}

static DBusMessage *adapter_get_fm_rdspower(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	const struct btd_adapter *adapter = data;
	DBusMessage *reply;
	
	uint8_t	reg_addr;
	uint8_t	get_value;	
	uint32_t ret =0;
	int dd;
	
	debug("%s:>>",__FUNCTION__);	
	
	if (!dbus_message_has_signature(msg, DBUS_TYPE_INVALID_AS_STRING))
		return invalid_args(msg);	

	reply = dbus_message_new_method_return(msg);
	if (!adapter->up)
		return adapter_not_ready(msg);

#ifndef DISABLE_STEP2

	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return no_such_adapter(msg);

	reg_addr	= BCM4325_I2C_FM_RDS_SYSTEM;
	
	if (bcm_vsc_read_reg(dd, reg_addr, &get_value, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}
	hci_close_dev(dd);
#endif	//DISABLE_STEP2

	//param  : 1 = on,   0 = off
	if((get_value & BCM4325_RDS_POWER_ON) == BCM4325_RDS_POWER_ON)
		ret = 1;	//ON
	else	
		ret = 0;	//OFF
	debug("		ret = %d", ret);		
	
	dbus_message_append_args(reply, DBUS_TYPE_UINT32, &ret,
					DBUS_TYPE_INVALID);	

	return reply;
}

static DBusMessage *adapter_set_fm_freq(DBusConnection *conn,
				DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;	//eastern modify
	//param mode : 1 = on,   0 = off
	uint8_t	reg_addr;
	uint32_t	set_value,get_value;
	uint32_t	freq;
	int dd;
	
	debug("%s:>>",__FUNCTION__);	
	
	if (!adapter->up)
		return adapter_not_ready(msg);

	if (!dbus_message_get_args(msg, NULL,
				DBUS_TYPE_UINT32, &set_value,
				DBUS_TYPE_INVALID))
		return invalid_args(msg);
	
	debug("set_value = %d 0x%X", set_value, set_value);	

#ifndef DISABLE_STEP2
	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return no_such_adapter(msg);
		

	set_value -= BCM4325_FREQUENCY_BASE;

	reg_addr	= BCM4325_I2C_FM_FREQ1;
	freq		= (set_value & 0xff00) >> 8;

	//write freq high byte
	if (bcm_vsc_write_reg(dd, reg_addr, (uint8_t)freq, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}

	reg_addr	= BCM4325_I2C_FM_FREQ0;
	freq		= set_value & 0x00ff;
	
	//write freq low byte
	if (bcm_vsc_write_reg(dd, reg_addr, (uint8_t)freq, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}

    //set tune mode
    set_value = BCM4325_FM_PRE_SET_MODE;
    if (bcm_vsc_write_reg(dd, BCM4325_I2C_FM_SEARCH_TUNE_MODE, (uint8_t)set_value, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}

	//if RDS power on, then enable interrupt.....
	//check RDS power
	if (bcm_vsc_read_reg(dd, BCM4325_I2C_FM_RDS_SYSTEM, &get_value, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}
	if((get_value & BCM4325_RDS_POWER_ON) == BCM4325_RDS_POWER_ON)
	{
	     //clear IRQ flag	
	     if ( (bcm_vsc_read_reg(dd, BCM4325_I2C_FM_RDS_FLAG0, &get_value, 1000) < 0) || 
		       (bcm_vsc_read_reg(dd, BCM4325_I2C_FM_RDS_FLAG1, &get_value, 1000) < 0) ) {
		     int err = errno;
		     error("hci_send_req() error: %s (%d)",
			     	strerror(errno), errno);
		     hci_close_dev(dd);
		  return failed_strerror(msg, err);
	     }

	     //enable water line interrupt
	     set_value = BCM4325_RDS_MASK_FIFO_WLINE;
	     if (bcm_vsc_write_reg(dd, BCM4325_I2C_FM_RDS_MASK1, (uint8_t)set_value, 1000) < 0) {
	    	int err = errno;
	     	error("hci_send_req() error: %s (%d)",
			    	strerror(errno), errno);
	    	hci_close_dev(dd);
		    return failed_strerror(msg, err);
	     }
	}
	hci_close_dev(dd);
#endif	//DISABLE_STEP2	
	return dbus_message_new_method_return(msg);
}

//[+][cellcore][Mingyan]2010/7/21 merge froyo
static DBusMessage *set_link_timeout(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;
        struct btd_device *device;
	const char *path;
	GSList *l;
	uint32_t num_slots;
        int dd, err;

	if (!dbus_message_get_args(msg, NULL,
			DBUS_TYPE_OBJECT_PATH, &path,
			DBUS_TYPE_UINT32, &num_slots,
			DBUS_TYPE_INVALID))
		return invalid_args(msg);

        l = g_slist_find_custom(adapter->devices,
                        path, (GCompareFunc) device_path_cmp);
        if (!l)
                return g_dbus_create_error(msg,
                                ERROR_INTERFACE ".DoesNotExist",
                                "Device does not exist");
	device = l->data;
	dd = hci_open_dev(adapter->dev_id);

	if (dd < 0) {
		err = -errno;
		goto fail;
	}

	err = hci_write_link_supervision_timeout(dd,
			htobs(device_get_handle(device)), htobs(num_slots), 1000);
	hci_close_dev(dd);

	if (err < 0) {
		err = -errno;
		goto fail;
	}
	return dbus_message_new_method_return(msg);
fail:
	return failed_strerror(msg, errno);
}
//[-][cellcore][Mingyan]2010/7/21 merge froyo

static DBusMessage *adapter_get_fm_freq(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	const struct btd_adapter *adapter = data;	//eastern modify
	DBusMessage *reply;
	
	uint8_t	reg_addr;
	uint8_t	get_value;	
	uint32_t	freq;
	int dd;
	
	debug("%s:>>",__FUNCTION__);	
	
	if (!dbus_message_has_signature(msg, DBUS_TYPE_INVALID_AS_STRING))
		return invalid_args(msg);	

	reply = dbus_message_new_method_return(msg);
	if (!adapter->up)
		return adapter_not_ready(msg);

#ifndef DISABLE_STEP2
	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return no_such_adapter(msg);

	
	//get high byte
	reg_addr	= BCM4325_I2C_FM_FREQ1;	//FREQ[15:8]
	if (bcm_vsc_read_reg(dd, reg_addr, &get_value, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}
	freq = get_value << 8;
	
    //get low byte
	reg_addr	= BCM4325_I2C_FM_FREQ0;	//FREQ[7:0]
	if (bcm_vsc_read_reg(dd, reg_addr, &get_value, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}

	freq = freq | get_value;
	freq = freq + BCM4325_FREQUENCY_BASE;


	hci_close_dev(dd);
#endif	//DISABLE_STEP2

	debug("		freq  = %d",freq );

	dbus_message_append_args(reply, DBUS_TYPE_UINT32, &freq,
					DBUS_TYPE_INVALID);	

	return reply;
}

static DBusMessage *adapter_set_fm_mute(DBusConnection *conn,
				DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;		//eastern 
	//param mode : 1= mute on   0=mute off
	uint8_t	reg_addr;
	uint32_t	set_value;
	int dd;
	
	debug("%s:>>",__FUNCTION__);	
	
	if (!adapter->up)
		return adapter_not_ready(msg);

	if (!dbus_message_get_args(msg, NULL,
				DBUS_TYPE_UINT32, &set_value,
				DBUS_TYPE_INVALID))
		return invalid_args(msg);
	
	debug("set_value = %d", set_value);	

#ifndef DISABLE_STEP2
	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return no_such_adapter(msg);


	reg_addr		= BCM4325_I2C_FM_AUDIO_CTRL0;
	if(0 == set_value)
	set_value	= (BCM4325_AUDIO_ROUTE_DAC | 
					BCM4325_DAC_OUTPUT_RIGHT | 
					BCM4325_DAC_OUTPUT_LEFT) ;	//0x1C
	else
	set_value	= (BCM4325_MANUAL_MUTE | 
					BCM4325_AUDIO_ROUTE_DAC | 
					BCM4325_DAC_OUTPUT_RIGHT | 
					BCM4325_DAC_OUTPUT_LEFT) ;	//0x1E
	
	if (bcm_vsc_write_reg(dd, reg_addr, (uint8_t)set_value, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}
	hci_close_dev(dd);
	
#endif	//DISABLE_STEP2	
	return dbus_message_new_method_return(msg);
}

static DBusMessage *adapter_get_fm_mute(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	const struct btd_adapter *adapter = data;		//eastern
	DBusMessage *reply;
	
	uint8_t	reg_addr;
	uint8_t	get_value;	
	uint32_t ret = 0;
	int dd;
	
	debug("%s:>>",__FUNCTION__);	
	
	if (!dbus_message_has_signature(msg, DBUS_TYPE_INVALID_AS_STRING))
		return invalid_args(msg);	

	reply = dbus_message_new_method_return(msg);
	if (!adapter->up)
		return adapter_not_ready(msg);

#ifndef DISABLE_STEP2

	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return no_such_adapter(msg);

	reg_addr	= BCM4325_I2C_FM_AUDIO_CTRL0;
	
	if (bcm_vsc_read_reg(dd, reg_addr, &get_value, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}	
	hci_close_dev(dd);
#endif	//DISABLE_STEP2

	//param  : 1 = on,   0 = off
	if((get_value & BCM4325_MANUAL_MUTE) == BCM4325_MANUAL_MUTE)
		ret = 1;	//ON
	else	
		ret = 0;	//OFF

	debug("		ret  = %d", ret);		
	
	dbus_message_append_args(reply, DBUS_TYPE_UINT32, &ret,
					DBUS_TYPE_INVALID);	

	return reply;
}

static DBusMessage *adapter_set_fm_search_step(DBusConnection *conn,
				DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;
	//param mode : 1= mute on   0=mute off
	uint8_t	reg_addr;
	uint32_t	set_value;
	int dd;
	
	debug("%s:>>",__FUNCTION__);	
	
	if (!adapter->up)
		return adapter_not_ready(msg);

	if (!dbus_message_get_args(msg, NULL,
				DBUS_TYPE_UINT32, &set_value,
				DBUS_TYPE_INVALID))
		return invalid_args(msg);
	
	debug("set_value = %d", set_value);	

#ifndef DISABLE_STEP2
	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return no_such_adapter(msg);

	reg_addr = BCM4325_I2C_FM_SEARCH_STEPS;
	
	//check step range 1~255
	if( 0 == set_value || set_value > 0xff )
	       set_value	= 0x64 ; // 100kHz
	
	if (bcm_vsc_write_reg(dd, reg_addr, (uint8_t)set_value, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}
	LOGE("   serach step = %d kHz",set_value);
	
	hci_close_dev(dd);
	
#endif	//DISABLE_STEP2	
	return dbus_message_new_method_return(msg);

}

static DBusMessage *adapter_start_fm_autosearch(DBusConnection *conn,
				DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;
	//param mode : 1= mute on   0=mute off
	uint8_t	reg_addr,get_value;
	uint32_t	set_value1,set_value2,freq,step;
	uint32_t ret;
	DBusMessage *reply;
	int dd;
	
	debug("%s:>>",__FUNCTION__);	
	
	if (!adapter->up)
		return adapter_not_ready(msg);

	if (!dbus_message_get_args(msg, NULL,
				DBUS_TYPE_UINT32, &set_value1,
				DBUS_TYPE_UINT32, &set_value2, 
				DBUS_TYPE_INVALID))
		return invalid_args(msg);
	
	debug("set_value1 = %d  set_value2= &d", set_value1,set_value2);	

#ifndef DISABLE_STEP2
	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return no_such_adapter(msg);
	/*indicate the starting frequency*/
	// offset from 64MHz
	set_value1 = set_value1 - BCM4325_FREQUENCY_BASE;

	//write freq high byte
	reg_addr	= BCM4325_I2C_FM_FREQ1;
	freq		= (set_value1 & 0xff00) >> 8;
	if (bcm_vsc_write_reg(dd, reg_addr, (uint8_t)freq, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}
	
	//write freq low byte
	reg_addr	= BCM4325_I2C_FM_FREQ0;
	freq		= set_value1 & 0x00ff;
	if (bcm_vsc_write_reg(dd, reg_addr, (uint8_t)freq, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}

	/*set RSSI threshold and search dir*/
	// bit 7 : 1= up 0=down
    if(0 == set_value2) 
    {
		//set_value2 = BCM4325_RSSI_LEVEL_BASE;  //-60dBm
		set_value2 = 0x5a; //-90dBm
	}
	else
	{
		//set_value2 = 0x80 | BCM4325_RSSI_LEVEL_BASE;
		set_value2 = 0x80 | 0x5a ;
	}
	if (bcm_vsc_write_reg(dd, BCM4325_I2C_FM_SEARCH_CTRL0, (uint8_t)set_value2, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}	

	//clear IRQ flag	
	if ( (bcm_vsc_read_reg(dd, BCM4325_I2C_FM_RDS_FLAG0, &get_value, 1000) < 0) || 
		  (bcm_vsc_read_reg(dd, BCM4325_I2C_FM_RDS_FLAG1, &get_value, 1000) < 0) ) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}

	//set auto search mode
	if (bcm_vsc_write_reg(dd, BCM4325_I2C_FM_SEARCH_TUNE_MODE, BCM4325_FM_AUTO_SEARCH_MODE, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}
	hci_close_dev(dd);
#endif	//DISABLE_STEP2

	return dbus_message_new_method_return(msg);
}

static DBusMessage *adapter_is_search_done(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	const struct btd_adapter *adapter = data;
	DBusMessage *reply;
	uint8_t	reg_addr;
	uint8_t	get_value;	
	uint32_t ret,isfinish;
	int dd;
	
	debug("%s:>>",__FUNCTION__);	
	
	if (!dbus_message_has_signature(msg, DBUS_TYPE_INVALID_AS_STRING))
		return invalid_args(msg);	

	reply = dbus_message_new_method_return(msg);
	if (!adapter->up)
		return adapter_not_ready(msg);

#ifndef DISABLE_STEP2

	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return no_such_adapter(msg);
    //Is search finish or fail
	reg_addr = BCM4325_I2C_FM_RDS_FLAG0;
	do {
	     if (bcm_vsc_read_reg(dd, reg_addr, &get_value, 1000) < 0) {
		    int err = errno;
		    error("hci_send_req() error: %s (%d)",strerror(errno), errno);
		    hci_close_dev(dd);
		    return failed_strerror(msg, err);
	    }
		isfinish = (get_value & BCM4325_FM_FLAG_SEARCH_TUNE_FINISHED);
	}while(isfinish==0);
	hci_close_dev(dd);
#endif	//DISABLE_STEP2

	if((get_value & BCM4325_FM_FLAG_SEARCH_TUNE_FAIL) == BCM4325_FM_FLAG_SEARCH_TUNE_FAIL)
		ret = 0;	//search fail
	else	
		ret = 1;	//search pass
		
	debug("		ret  = %d", ret);	
	dbus_message_append_args(reply, DBUS_TYPE_UINT32, &ret,
					DBUS_TYPE_INVALID);	
	return reply;
}

static DBusMessage *adapter_set_fm_monostereo(DBusConnection *conn,
				DBusMessage *msg, void *data)
{
	struct btd_adapter *adapter = data;
	uint8_t	reg_addr;
	uint32_t	set_value;
	int dd;
	
	debug("%s:>>",__FUNCTION__);	
	
	if (!adapter->up)
		return adapter_not_ready(msg);

	if (!dbus_message_get_args(msg, NULL,
				DBUS_TYPE_UINT32, &set_value,
				DBUS_TYPE_INVALID))
		return invalid_args(msg);
	
	debug("set_value = %d", set_value);	

#ifndef DISABLE_STEP2
	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return no_such_adapter(msg);

	reg_addr		= BCM4325_I2C_FM_CTRL;
	//param mode : 1 = stereo, 0=mono
	if( 0 == set_value)
	set_value	= 0x00;
	else
	set_value	= BCM4325_STEREO_MONO_MANUAL_SELECT;
	
	if (bcm_vsc_write_reg(dd, reg_addr, (uint8_t)set_value, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}
	
	hci_close_dev(dd);
#endif	//DISABLE_STEP2

	return dbus_message_new_method_return(msg);
}

static DBusMessage *adapter_get_fm_monostereo(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	const struct btd_adapter *adapter = data;
	DBusMessage *reply;
	
	uint8_t	reg_addr;
	uint8_t	get_value;	
	uint32_t ret = 0 ;
	int dd;
	
	debug("%s:>>",__FUNCTION__);	
	
	if (!dbus_message_has_signature(msg, DBUS_TYPE_INVALID_AS_STRING))
		return invalid_args(msg);	

	reply = dbus_message_new_method_return(msg);
	if (!adapter->up)
		return adapter_not_ready(msg);

#ifndef DISABLE_STEP2
	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return no_such_adapter(msg);

	reg_addr = BCM4325_I2C_FM_RDS_FLAG0;
	
	if (bcm_vsc_read_reg(dd, reg_addr, &get_value, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}
	hci_close_dev(dd);
#endif	//DISABLE_STEP2

	//param  : 1 = stereo,  0 = mono
	if((get_value & BCM4325_FLAG_STEREO_DETECTED) == BCM4325_FLAG_STEREO_DETECTED)
		ret = 1;	//stereo
	else	
		ret = 0;	//mono
		
	debug("		ret  = %d", ret);	
	dbus_message_append_args(reply, DBUS_TYPE_UINT32, &ret,
					DBUS_TYPE_INVALID);	
	return reply;
}

static DBusMessage *adapter_get_fm_rdsdata_psn(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	const struct btd_adapter *adapter = data;
	DBusMessage *reply;
	DBusMessageIter iter, array_iter;
	
	uint8_t	reg_addr;
	uint8_t	get_value,set_value;
	uint8_t	rds_data[BCM4325_RDS_FIFO_WLINE*BCM4325_RDS_FIFO_DUPLE_SIZE];	
	uint8_t rds_ps[BCM4325_RDS_PSN_LENGTH+1],*ptr_rds_ps = rds_ps;
	uint32_t ret;
	int dd;
	//int i;
	
	debug("%s:>>",__FUNCTION__);	
	
	if (!dbus_message_has_signature(msg, DBUS_TYPE_INVALID_AS_STRING))
		return invalid_args(msg);	

	reply = dbus_message_new_method_return(msg);
	if (!adapter->up)
		return adapter_not_ready(msg);

#ifndef DISABLE_STEP2
	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return no_such_adapter(msg);

	reg_addr = BCM4325_I2C_RDS_DATA;
	//get RDS data
	//rds_data size 150 byte
	if (bcm_vsc_read_rds_reg(dd, reg_addr, rds_data, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}

	//for(i=0; i<BCM4325_RDS_FIFO_WLINE; i++)
	//{
	//	LOGI("RDS_data[%d] : 0x%02x  0x%02x  0x%02x",i,rds_data[3*i],rds_data[3*i+1],rds_data[3*i+2]);
	//}
	
    //get program service name
	BCM4325_updateRDSInfo(rds_data);
	BCM4325_getPSName(rds_ps);

	LOGI("GetPSName %c%c%c%c%c%c%c%c \r\n",
         rds_ps[0],rds_ps[1],rds_ps[2],rds_ps[3],
         rds_ps[4],rds_ps[5],rds_ps[6],rds_ps[7]);
	
    //clear IRQ flag	
	if ( (bcm_vsc_read_reg(dd, BCM4325_I2C_FM_RDS_FLAG0, &get_value, 1000) < 0) || 
		  (bcm_vsc_read_reg(dd, BCM4325_I2C_FM_RDS_FLAG1, &get_value, 1000) < 0) ) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}

	//enable water line interrupt
	set_value = BCM4325_RDS_MASK_FIFO_WLINE;
	if (bcm_vsc_write_reg(dd, BCM4325_I2C_FM_RDS_MASK1, (uint8_t)set_value, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}
	
	hci_close_dev(dd);
#endif	//DISABLE_STEP2

   	dbus_message_iter_init_append(reply, &iter);
	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
				DBUS_TYPE_BYTE_AS_STRING, &array_iter);

	dbus_message_iter_append_fixed_array(&array_iter,
				DBUS_TYPE_BYTE, &ptr_rds_ps, sizeof(rds_ps));

	dbus_message_iter_close_container(&iter, &array_iter);

	return reply;
}

static DBusMessage *adapter_get_fm_rssi(DBusConnection *conn,
					DBusMessage *msg, void *data)
{
	const struct btd_adapter *adapter = data;
	DBusMessage *reply;
	
	uint8_t	reg_addr;
	uint8_t	get_value;	
	uint32_t ret;
	int dd;
	
	debug("%s:>>",__FUNCTION__);	
	
	if (!dbus_message_has_signature(msg, DBUS_TYPE_INVALID_AS_STRING))
		return invalid_args(msg);	

	reply = dbus_message_new_method_return(msg);
	if (!adapter->up)
		return adapter_not_ready(msg);

#ifndef DISABLE_STEP2
	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return no_such_adapter(msg);

	reg_addr = BCM4325_I2C_FM_RSSI;
	
	if (bcm_vsc_read_reg(dd, reg_addr, &get_value, 1000) < 0) {
		int err = errno;
		error("hci_send_req() error: %s (%d)",
				strerror(errno), errno);
		hci_close_dev(dd);
		return failed_strerror(msg, err);
	}
	hci_close_dev(dd);
#endif	//DISABLE_STEP2
	
	if( get_value > 127 )
		ret = 256 - get_value;
	else	
		ret = get_value;
		
	debug("		ret  = %d", ret);	
	dbus_message_append_args(reply, DBUS_TYPE_UINT32, &ret,
					DBUS_TYPE_INVALID);	
	return reply;
}
//#endif //CONFIG_GEMINI_GENERIC

static GDBusMethodTable adapter_methods[] = {
	{ "GetProperties",	"",	"a{sv}",get_properties		},
	{ "SetProperty",	"sv",	"",	set_property,
						G_DBUS_METHOD_FLAG_ASYNC},
	{ "RequestSession",	"",	"",	request_session,
						G_DBUS_METHOD_FLAG_ASYNC},
	{ "ReleaseSession",	"",	"",	release_session		},
	{ "StartDiscovery",	"",	"",	adapter_start_discovery },
	{ "StopDiscovery",	"",	"",	adapter_stop_discovery,
						G_DBUS_METHOD_FLAG_ASYNC},
	{ "ListDevices",	"",	"ao",	list_devices,
						G_DBUS_METHOD_FLAG_DEPRECATED},
	{ "CreateDevice",	"s",	"o",	create_device,
						G_DBUS_METHOD_FLAG_ASYNC},
	{ "CreatePairedDevice",	"sos",	"o",	create_paired_device,
						G_DBUS_METHOD_FLAG_ASYNC},
	{ "CancelDeviceCreation","s",	"",	cancel_device_creation,
						G_DBUS_METHOD_FLAG_ASYNC},
	{ "RemoveDevice",	"o",	"",	remove_device,
						G_DBUS_METHOD_FLAG_ASYNC},
	{ "FindDevice",		"s",	"o",	find_device		},
	{ "RegisterAgent",	"os",	"",	register_agent		},
	{ "UnregisterAgent",	"o",	"",	unregister_agent	},
	{ "AddRfcommServiceRecord",	"sttq",	"u",	add_rfcomm_service_record },
	{ "RemoveServiceRecord",	"u",	"",	remove_service_record },
	{ "SetLinkTimeout",	"ou",	"",	set_link_timeout	},
//#ifdef CONFIG_GEMINI_GENERIC
	{ "setFmPower",			"u",	"",						adapter_set_fm_power		},						
	{ "getFmPower",			"",	"u",						adapter_get_fm_power		},
	{ "setFmRDSPower",		"u",	"",					adapter_set_fm_rdspower		},						
	{ "getFmRDSPower",		"",	"u",  				adapter_get_fm_rdspower		},				
	{ "setFmFreq",			"u",	"",						adapter_set_fm_freq		},						
	{ "getFmFreq",			"",	"u",						adapter_get_fm_freq		},		
	{ "setFmMute",			"u",	"",						adapter_set_fm_mute		},						
	{ "getFmMute",			"",	"u",						adapter_get_fm_mute		},								
	{ "setFmSearchStep",			"u",	"",			adapter_set_fm_search_step		},					
	{ "startFmAutoSearch",		"uu",	"",			adapter_start_fm_autosearch		},
	{ "isSearchDone",			"",	"u",					adapter_is_search_done		},						
        { "setFmMonoStereo",		"u",	"",     adapter_set_fm_monostereo       },
        { "getFmMonoStereo",		"",	    "u",  adapter_get_fm_monostereo       },                        
        { "getFmRDSData_PS",		"",	"ay",     adapter_get_fm_rdsdata_psn      },
	{ "getFmRSSI",			"",	"u",						adapter_get_fm_rssi		},	                        
//#endif						
	{ }
};

static GDBusSignalTable adapter_signals[] = {
	{ "PropertyChanged",		"sv"		},
	{ "DeviceCreated",		"o"		},
	{ "DeviceRemoved",		"o"		},
	{ "DeviceFound",		"sa{sv}"	},
	{ "DeviceDisappeared",		"s"		},
	{ }
};

static inline uint8_t get_inquiry_mode(struct hci_dev *dev)
{
	if (dev->features[6] & LMP_EXT_INQ)
		return 2;

	if (dev->features[3] & LMP_RSSI_INQ)
		return 1;

	if (dev->manufacturer == 11 &&
			dev->hci_rev == 0x00 && dev->lmp_subver == 0x0757)
		return 1;

	if (dev->manufacturer == 15) {
		if (dev->hci_rev == 0x03 && dev->lmp_subver == 0x6963)
			return 1;
		if (dev->hci_rev == 0x09 && dev->lmp_subver == 0x6963)
			return 1;
		if (dev->hci_rev == 0x00 && dev->lmp_subver == 0x6965)
			return 1;
	}

	if (dev->manufacturer == 31 &&
			dev->hci_rev == 0x2005 && dev->lmp_subver == 0x1805)
		return 1;

	return 0;
}

static int adapter_read_bdaddr(uint16_t dev_id, bdaddr_t *bdaddr)
{
	int dd, err;

	dd = hci_open_dev(dev_id);
	if (dd < 0) {
		err = -errno;
		error("Can't open device hci%d: %s (%d)",
					dev_id, strerror(errno), errno);
		return err;
	}

	if (hci_read_bd_addr(dd, bdaddr, HCI_REQ_TIMEOUT) < 0) {
		err = -errno;
		error("Can't read address for hci%d: %s (%d)",
					dev_id, strerror(errno), errno);
		hci_close_dev(dd);
		return err;
	}

	hci_close_dev(dd);

	return 0;
}

static int adapter_setup(struct btd_adapter *adapter)
{
	struct hci_dev *dev = &adapter->dev;
	uint8_t events[8] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0x1f, 0x00, 0x00 };
	uint8_t inqmode;
	int err , dd;
	char name[MAX_NAME_LENGTH + 1];

	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0) {
		err = -errno;
		error("Can't open device hci%d: %s (%d)", adapter->dev_id,
						strerror(errno), errno);
		return err;
	}

	if (dev->lmp_ver > 1) {
		if (dev->features[5] & LMP_SNIFF_SUBR)
			events[5] |= 0x20;

		if (dev->features[5] & LMP_PAUSE_ENC)
			events[5] |= 0x80;

		if (dev->features[6] & LMP_EXT_INQ)
			events[5] |= 0x40;

		if (dev->features[6] & LMP_NFLUSH_PKTS)
			events[7] |= 0x01;

		if (dev->features[7] & LMP_LSTO)
			events[6] |= 0x80;

		if (dev->features[6] & LMP_SIMPLE_PAIR) {
			events[6] |= 0x01;	/* IO Capability Request */
			events[6] |= 0x02;	/* IO Capability Response */
			events[6] |= 0x04;	/* User Confirmation Request */
			events[6] |= 0x08;	/* User Passkey Request */
			events[6] |= 0x10;	/* Remote OOB Data Request */
			events[6] |= 0x20;	/* Simple Pairing Complete */
			events[7] |= 0x04;	/* User Passkey Notification */
			events[7] |= 0x08;	/* Keypress Notification */
			events[7] |= 0x10;	/* Remote Host Supported
						 * Features Notification */
		}

		hci_send_cmd(dd, OGF_HOST_CTL, OCF_SET_EVENT_MASK,
						sizeof(events), events);
	}

	if (read_local_name(&adapter->bdaddr, name) == 0)
		adapter_ops->set_name(adapter->dev_id, name);

	inqmode = get_inquiry_mode(dev);
	if (inqmode < 1)
		goto done;

	if (hci_write_inquiry_mode(dd, inqmode, HCI_REQ_TIMEOUT) < 0) {
		err = -errno;
		error("Can't write inquiry mode for %s: %s (%d)",
					adapter->path, strerror(errno), errno);
		hci_close_dev(dd);
		return err;
	}

done:
	hci_close_dev(dd);
	return 0;
}

static void create_stored_device_from_profiles(char *key, char *value,
						void *user_data)
{
	struct btd_adapter *adapter = user_data;
	GSList *uuids = bt_string2list(value);
	struct btd_device *device;
	bdaddr_t dst;
	char srcaddr[18], dstaddr[18];

	ba2str(&adapter->bdaddr, srcaddr);

	if (g_slist_find_custom(adapter->devices,
				key, (GCompareFunc) device_address_cmp))
		return;

	device = device_create(connection, adapter, key);
	if (!device)
		return;

	device_set_temporary(device, FALSE);
	adapter->devices = g_slist_append(adapter->devices, device);

	device_get_address(device, &dst);
	ba2str(&dst, dstaddr);

	device_probe_drivers(device, uuids);

	g_slist_foreach(uuids, (GFunc) g_free, NULL);
	g_slist_free(uuids);
}

static void create_stored_device_from_linkkeys(char *key, char *value,
						void *user_data)
{
	struct btd_adapter *adapter = user_data;
	struct btd_device *device;

	if (g_slist_find_custom(adapter->devices,
				key, (GCompareFunc) device_address_cmp))
		return;

	device = device_create(connection, adapter, key);
	if (device) {
		device_set_temporary(device, FALSE);
		adapter->devices = g_slist_append(adapter->devices, device);
	}
}

static void load_devices(struct btd_adapter *adapter)
{
	char filename[PATH_MAX + 1];
	char srcaddr[18];

	ba2str(&adapter->bdaddr, srcaddr);

	create_name(filename, PATH_MAX, STORAGEDIR, srcaddr, "profiles");
	textfile_foreach(filename, create_stored_device_from_profiles,
								adapter);

	create_name(filename, PATH_MAX, STORAGEDIR, srcaddr, "linkkeys");
	textfile_foreach(filename, create_stored_device_from_linkkeys,
								adapter);
}

static void probe_driver(gpointer data, gpointer user_data)
{
	struct btd_adapter *adapter = data;
	struct btd_adapter_driver *driver = user_data;
	int err;

	if (!adapter->up)
		return;

	err = driver->probe(adapter);
	if (err < 0)
		error("%s: %s (%d)", driver->name, strerror(-err), -err);
}

static void load_drivers(struct btd_adapter *adapter)
{
	GSList *l;

	for (l = adapter_drivers; l; l = l->next) {
		struct btd_adapter_driver *driver = l->data;

		if (driver->probe == NULL)
			continue;

		probe_driver(adapter, driver);
	}
}

static void load_connections(struct btd_adapter *adapter)
{
	struct hci_conn_list_req *cl = NULL;
	struct hci_conn_info *ci;
	int i, dd;

	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0)
		return;

	cl = g_malloc0(10 * sizeof(*ci) + sizeof(*cl));

	cl->dev_id = adapter->dev_id;
	cl->conn_num = 10;
	ci = cl->conn_info;

	if (ioctl(dd, HCIGETCONNLIST, cl) != 0) {
		g_free(cl);
		hci_close_dev(dd);
		return;
	}

	for (i = 0; i < cl->conn_num; i++, ci++) {
		struct btd_device *device;
		char address[18];

		ba2str(&ci->bdaddr, address);
		device = adapter_get_device(connection, adapter, address);
		if (device)
			adapter_add_connection(adapter, device, ci->handle);
	}

	g_free(cl);
	hci_close_dev(dd);
}

static int get_discoverable_timeout(const char *src)
{
	int timeout;

	if (read_discoverable_timeout(src, &timeout) == 0)
		return timeout;

	return main_opts.discovto;
}

static int get_pairable_timeout(const char *src)
{
	int timeout;

	if (read_pairable_timeout(src, &timeout) == 0)
		return timeout;

	return main_opts.pairto;
}

static int adapter_up(struct btd_adapter *adapter)
{
	char mode[14], srcaddr[18];
	uint8_t scan_mode;
	gboolean powered, dev_down = FALSE;
	int err;

	ba2str(&adapter->bdaddr, srcaddr);

	adapter->off_requested = FALSE;
	adapter->up = 1;
	adapter->discov_timeout = get_discoverable_timeout(srcaddr);
	adapter->pairable_timeout = get_pairable_timeout(srcaddr);
	adapter->state = DISCOVER_TYPE_NONE;
	adapter->mode = MODE_CONNECTABLE;
	adapter->cache_enable = TRUE;
	scan_mode = SCAN_PAGE;
	powered = TRUE;

	/* Set pairable mode */
	if (read_device_pairable(&adapter->bdaddr, &adapter->pairable) < 0)
		adapter->pairable = TRUE;

	if (!adapter->initialized && !main_opts.remember_powered) {
		if (main_opts.mode == MODE_OFF)
			strcpy(mode, "off");
		else
			strcpy(mode, "connectable");
	} else if (read_device_mode(srcaddr, mode, sizeof(mode)) < 0) {
		if (!adapter->initialized && main_opts.mode == MODE_OFF)
			strcpy(mode, "off");
		else
			goto proceed;
	}

	if (g_str_equal(mode, "off")) {
		powered = FALSE;

		if (!adapter->initialized) {
			dev_down = TRUE;
			goto proceed;
		}

		if (read_on_mode(srcaddr, mode, sizeof(mode)) < 0 ||
						g_str_equal(mode, "off"))
			write_device_mode(&adapter->bdaddr, "connectable");
		else
			write_device_mode(&adapter->bdaddr, mode);

		return adapter_up(adapter);
	} else if (!g_str_equal(mode, "connectable") &&
			adapter->discov_timeout == 0) {
		/* Set discoverable only if timeout is 0 */
		adapter->mode = MODE_DISCOVERABLE;
		scan_mode = SCAN_PAGE | SCAN_INQUIRY;
	}

proceed:
	if (scan_mode == SCAN_PAGE)
		err = adapter_ops->set_connectable(adapter->dev_id);
	else
		err = adapter_ops->set_discoverable(adapter->dev_id);

	if (err < 0)
		return err;

	if (adapter->initialized == FALSE) {
		load_drivers(adapter);
		load_devices(adapter);

		/* retrieve the active connections: address the scenario where
		 * the are active connections before the daemon've started */
		load_connections(adapter);

		adapter->initialized = TRUE;

		manager_add_adapter(adapter->path);

	}

	if (dev_down) {
		adapter_ops->stop(adapter->dev_id);
		adapter->off_requested = TRUE;
		return 1;
	} else
		emit_property_changed(connection, adapter->path,
					ADAPTER_INTERFACE, "Powered",
					DBUS_TYPE_BOOLEAN, &powered);

	adapter_disable_svc_cache(adapter);
	return 0;
}

int adapter_start(struct btd_adapter *adapter)
{
	struct hci_dev *dev = &adapter->dev;
	struct hci_dev_info di;
	struct hci_version ver;
	uint8_t features[8];
	int dd, err;

	if (hci_devinfo(adapter->dev_id, &di) < 0)
		return -errno;

	if (hci_test_bit(HCI_RAW, &di.flags)) {
		dev->ignore = 1;
		return -1;
	}

	if (!bacmp(&di.bdaddr, BDADDR_ANY)) {
		int err;

		debug("Adapter %s without an address", adapter->path);

		err = adapter_read_bdaddr(adapter->dev_id, &di.bdaddr);
		if (err < 0)
			return err;
	}

	bacpy(&adapter->bdaddr, &di.bdaddr);
	memcpy(dev->features, di.features, 8);

	dd = hci_open_dev(adapter->dev_id);
	if (dd < 0) {
		err = -errno;
		error("Can't open adapter %s: %s (%d)",
					adapter->path, strerror(errno), errno);
		return err;
	}

	if (hci_read_local_version(dd, &ver, HCI_REQ_TIMEOUT) < 0) {
		err = -errno;
		error("Can't read version info for %s: %s (%d)",
					adapter->path, strerror(errno), errno);
		hci_close_dev(dd);
		return err;
	}

	dev->hci_rev = ver.hci_rev;
	dev->lmp_ver = ver.lmp_ver;
	dev->lmp_subver = ver.lmp_subver;
	dev->manufacturer = ver.manufacturer;

	if (hci_read_local_features(dd, features, HCI_REQ_TIMEOUT) < 0) {
		err = -errno;
		error("Can't read features for %s: %s (%d)",
					adapter->path, strerror(errno), errno);
		hci_close_dev(dd);
		return err;
	}

	memcpy(dev->features, features, 8);

	if (hci_read_class_of_dev(dd, dev->class, HCI_REQ_TIMEOUT) < 0) {
		err = -errno;
		error("Can't read class of adapter on %s: %s (%d)",
					adapter->path, strerror(errno), errno);
		hci_close_dev(dd);
		return err;
	}

	adapter_ops->read_name(adapter->dev_id);

	if (!(features[6] & LMP_SIMPLE_PAIR))
		goto setup;

	if (ioctl(dd, HCIGETAUTHINFO, NULL) < 0 && errno != EINVAL)
		hci_write_simple_pairing_mode(dd, 0x01, HCI_REQ_TIMEOUT);

	if (hci_read_simple_pairing_mode(dd, &dev->ssp_mode,
						HCI_REQ_TIMEOUT) < 0) {
		err = -errno;
		error("Can't read simple pairing mode on %s: %s (%d)",
					adapter->path, strerror(errno), errno);
		/* Fall through since some chips have broken
		 * read_simple_pairing_mode behavior */
	}

setup:
	hci_send_cmd(dd, OGF_LINK_POLICY, OCF_READ_DEFAULT_LINK_POLICY,
								0, NULL);
	hci_close_dev(dd);

	adapter_setup(adapter);

	if (!adapter->initialized && adapter->already_up) {
		debug("Stopping Inquiry at adapter startup");
		adapter_ops->stop_discovery(adapter->dev_id);
	}

	err = adapter_up(adapter);

	info("Adapter %s has been enabled", adapter->path);

	return err;
}

static void reply_pending_requests(struct btd_adapter *adapter)
{
	GSList *l;

	if (!adapter)
		return;

	/* pending bonding */
	for (l = adapter->devices; l; l = l->next) {
		struct btd_device *device = l->data;

		if (device_is_bonding(device, NULL))
			device_cancel_bonding(device,
						HCI_OE_USER_ENDED_CONNECTION);
	}

	if (adapter->state & STD_INQUIRY || adapter->state & PERIODIC_INQUIRY) {
		/* Cancel inquiry initiated by D-Bus client */
		if (adapter->disc_sessions)
			adapter_ops->stop_discovery(adapter->dev_id);
	}
}

static void unload_drivers(struct btd_adapter *adapter)
{
	GSList *l;

	for (l = adapter_drivers; l; l = l->next) {
		struct btd_adapter_driver *driver = l->data;

		if (driver->remove)
			driver->remove(adapter);
	}
}

int adapter_stop(struct btd_adapter *adapter)
{
	gboolean powered, discoverable, pairable;

	/* cancel pending timeout */
	if (adapter->discov_timeout_id) {
		g_source_remove(adapter->discov_timeout_id);
		adapter->discov_timeout_id = 0;
	}

	/* check pending requests */
	reply_pending_requests(adapter);

	if (adapter->disc_sessions) {
		g_slist_foreach(adapter->disc_sessions, (GFunc) session_free,
				NULL);
		g_slist_free(adapter->disc_sessions);
		adapter->disc_sessions = NULL;
	}

	clear_found_devices_list(adapter);

	if (adapter->oor_devices) {
		g_slist_free(adapter->oor_devices);
		adapter->oor_devices = NULL;
	}

	while (adapter->connections) {
		struct btd_device *device = adapter->connections->data;
		adapter_remove_connection(adapter, device, 0);
	}

	if (adapter->scan_mode == (SCAN_PAGE | SCAN_INQUIRY)) {
		discoverable = FALSE;
		emit_property_changed(connection, adapter->path,
					ADAPTER_INTERFACE, "Discoverable",
					DBUS_TYPE_BOOLEAN, &discoverable);
	}

	if ((adapter->scan_mode & SCAN_PAGE) && adapter->pairable == TRUE) {
		pairable = FALSE;
		emit_property_changed(connection, adapter->path,
					ADAPTER_INTERFACE, "Pairable",
					DBUS_TYPE_BOOLEAN, &pairable);
	}

	powered = FALSE;
	emit_property_changed(connection, adapter->path, ADAPTER_INTERFACE,
				"Powered", DBUS_TYPE_BOOLEAN, &powered);

	adapter->up = 0;
	adapter->scan_mode = SCAN_DISABLED;
	adapter->mode = MODE_OFF;
	adapter->state = DISCOVER_TYPE_NONE;
	adapter->cache_enable = TRUE;

	info("Adapter %s has been disabled", adapter->path);

	return 0;
}

int adapter_update(struct btd_adapter *adapter, uint8_t new_svc)
{
	struct hci_dev *dev = &adapter->dev;

	if (dev->ignore)
		return 0;

	if (adapter->cache_enable) {
		adapter->svc_cache = new_svc;
		return 0;
	}

	set_service_classes(adapter, new_svc);

	update_ext_inquiry_response(adapter);

	return 0;
}

void adapter_disable_svc_cache(struct btd_adapter *adapter)
{
	if (!adapter)
		return;

	if (!adapter->cache_enable)
		return;

	/* Disable and flush svc cache. All successive service class updates
	   will be written to the device */
	adapter->cache_enable = FALSE;

	set_service_classes(adapter, adapter->svc_cache);

	update_ext_inquiry_response(adapter);
}

int adapter_get_class(struct btd_adapter *adapter, uint8_t *cls)
{
	struct hci_dev *dev = &adapter->dev;

	memcpy(cls, dev->class, 3);

	return 0;
}

int adapter_set_class(struct btd_adapter *adapter, uint8_t *cls)
{
	struct hci_dev *dev = &adapter->dev;
	uint32_t class;

	if (memcmp(dev->class, cls, 3) == 0)
		return 0;

	memcpy(dev->class, cls, 3);

	write_local_class(&adapter->bdaddr, cls);

	class = cls[0] | (cls[1] << 8) | (cls[2] << 16);

	emit_property_changed(connection, adapter->path, ADAPTER_INTERFACE,
				"Class", DBUS_TYPE_UINT32, &class);

	return 0;
}

int adapter_update_ssp_mode(struct btd_adapter *adapter, uint8_t mode)
{
	struct hci_dev *dev = &adapter->dev;

	dev->ssp_mode = mode;

	update_ext_inquiry_response(adapter);

	return 0;
}

static void adapter_free(gpointer user_data)
{
	struct btd_adapter *adapter = user_data;

	agent_destroy(adapter->agent, FALSE);
	adapter->agent = NULL;

	debug("adapter_free(%p)", adapter);

	g_free(adapter->path);
	g_free(adapter);
}

struct btd_adapter *btd_adapter_ref(struct btd_adapter *adapter)
{
	adapter->ref++;

	debug("btd_adapter_ref(%p): ref=%d", adapter, adapter->ref);

	return adapter;
}

void btd_adapter_unref(struct btd_adapter *adapter)
{
	gchar *path;

	adapter->ref--;

	debug("btd_adapter_unref(%p): ref=%d", adapter, adapter->ref);

	if (adapter->ref > 0)
		return;

	path = g_strdup(adapter->path);

	g_dbus_unregister_interface(connection, path, ADAPTER_INTERFACE);

	g_free(path);
}

struct btd_adapter *adapter_create(DBusConnection *conn, int id,
				gboolean devup)
{
	char path[MAX_PATH_LENGTH];
	struct btd_adapter *adapter;
	const char *base_path = manager_get_base_path();

	if (!connection)
		connection = conn;

	snprintf(path, sizeof(path), "%s/hci%d", base_path, id);

	adapter = g_try_new0(struct btd_adapter, 1);
	if (!adapter) {
		error("adapter_create: failed to alloc memory for %s", path);
		return NULL;
	}

	adapter->dev_id = id;
	if (main_opts.name_resolv)
		adapter->state |= RESOLVE_NAME;
	adapter->path = g_strdup(path);
	adapter->already_up = devup;

	if (!g_dbus_register_interface(conn, path, ADAPTER_INTERFACE,
			adapter_methods, adapter_signals, NULL,
			adapter, adapter_free)) {
		error("Adapter interface init failed on path %s", path);
		adapter_free(adapter);
		return NULL;
	}

	return btd_adapter_ref(adapter);
}

void adapter_remove(struct btd_adapter *adapter)
{
	GSList *l;

	debug("Removing adapter %s", adapter->path);

	for (l = adapter->devices; l; l = l->next)
		device_remove(l->data, connection, FALSE);
	g_slist_free(adapter->devices);

	unload_drivers(adapter);

	/* Return adapter to down state if it was not up on init */
	if (adapter->up && !adapter->already_up)
		adapter_ops->stop(adapter->dev_id);

	btd_adapter_unref(adapter);
}

uint16_t adapter_get_dev_id(struct btd_adapter *adapter)
{
	return adapter->dev_id;
}

const gchar *adapter_get_path(struct btd_adapter *adapter)
{
	if (!adapter)
		return NULL;

	return adapter->path;
}

void adapter_get_address(struct btd_adapter *adapter, bdaddr_t *bdaddr)
{
	bacpy(bdaddr, &adapter->bdaddr);
}

void adapter_set_state(struct btd_adapter *adapter, int state)
{
	gboolean discov_active = FALSE;
	const char *path = adapter->path;

	if (adapter->state == state)
		return;

	if (state & PERIODIC_INQUIRY || state & STD_INQUIRY)
		discov_active = TRUE;
	else if (adapter->disc_sessions && main_opts.discov_interval)
		adapter->scheduler_id = g_timeout_add_seconds(
						main_opts.discov_interval,
						(GSourceFunc) adapter_start_inquiry,
						adapter);

	/* Send out of range */
	if (!discov_active)
		adapter_update_oor_devices(adapter);

	emit_property_changed(connection, path,
				ADAPTER_INTERFACE, "Discovering",
				DBUS_TYPE_BOOLEAN, &discov_active);

	adapter->state = state;
}

int adapter_get_state(struct btd_adapter *adapter)
{
	return adapter->state;
}

gboolean adapter_is_ready(struct btd_adapter *adapter)
{
	return adapter->initialized;
}

struct remote_dev_info *adapter_search_found_devices(struct btd_adapter *adapter,
						struct remote_dev_info *match)
{
	GSList *l;

	l = g_slist_find_custom(adapter->found_devices, match,
					(GCompareFunc) found_device_cmp);
	if (l)
		return l->data;

	return NULL;
}

static int dev_rssi_cmp(struct remote_dev_info *d1, struct remote_dev_info *d2)
{
	int rssi1, rssi2;

	rssi1 = d1->rssi < 0 ? -d1->rssi : d1->rssi;
	rssi2 = d2->rssi < 0 ? -d2->rssi : d2->rssi;

	return rssi1 - rssi2;
}

static void append_dict_valist(DBusMessageIter *iter,
					const char *first_key,
					va_list var_args)
{
	DBusMessageIter dict;
	const char *key;
	int type;
	void *val;

	dbus_message_iter_open_container(iter, DBUS_TYPE_ARRAY,
			DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);

	key = first_key;
	while (key) {
		type = va_arg(var_args, int);
		val = va_arg(var_args, void *);
		dict_append_entry(&dict, key, type, val);
		key = va_arg(var_args, char *);
	}

	dbus_message_iter_close_container(iter, &dict);
}

static void emit_device_found(const char *path, const char *address,
				const char *first_key, ...)
{
	DBusMessage *signal;
	DBusMessageIter iter;
	va_list var_args;

	signal = dbus_message_new_signal(path, ADAPTER_INTERFACE,
					"DeviceFound");
	if (!signal) {
		error("Unable to allocate new %s.DeviceFound signal",
				ADAPTER_INTERFACE);
		return;
	}
	dbus_message_iter_init_append(signal, &iter);
	dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &address);

	va_start(var_args, first_key);
	append_dict_valist(&iter, first_key, var_args);
	va_end(var_args);

	g_dbus_send_message(connection, signal);
}

void adapter_emit_device_found(struct btd_adapter *adapter,
				struct remote_dev_info *dev)
{
	struct btd_device *device;
	char peer_addr[18], local_addr[18];
	const char *icon, *paddr = peer_addr;
	dbus_bool_t paired = FALSE;
	dbus_int16_t rssi = dev->rssi;
	char *alias;

	ba2str(&dev->bdaddr, peer_addr);
	ba2str(&adapter->bdaddr, local_addr);

	device = adapter_find_device(adapter, paddr);
	if (device)
		paired = device_is_paired(device);

	icon = class_to_icon(dev->class);

	if (!dev->alias) {
		if (!dev->name) {
			alias = g_strdup(peer_addr);
			g_strdelimit(alias, ":", '-');
		} else
			alias = g_strdup(dev->name);
	} else
		alias = g_strdup(dev->alias);

	emit_device_found(adapter->path, paddr,
			"Address", DBUS_TYPE_STRING, &paddr,
			"Class", DBUS_TYPE_UINT32, &dev->class,
			"Icon", DBUS_TYPE_STRING, &icon,
			"RSSI", DBUS_TYPE_INT16, &rssi,
			"Name", DBUS_TYPE_STRING, &dev->name,
			"Alias", DBUS_TYPE_STRING, &alias,
			"LegacyPairing", DBUS_TYPE_BOOLEAN, &dev->legacy,
			"Paired", DBUS_TYPE_BOOLEAN, &paired,
			NULL);

	g_free(alias);
}

void adapter_update_found_devices(struct btd_adapter *adapter, bdaddr_t *bdaddr,
				int8_t rssi, uint32_t class, const char *name,
				const char *alias, gboolean legacy,
				name_status_t name_status)
{
	struct remote_dev_info *dev, match;

	memset(&match, 0, sizeof(struct remote_dev_info));
	bacpy(&match.bdaddr, bdaddr);
	match.name_status = NAME_ANY;

	dev = adapter_search_found_devices(adapter, &match);
	if (dev) {
		if (rssi == dev->rssi)
			return;

		/* Out of range list update */
		adapter->oor_devices = g_slist_remove(adapter->oor_devices,
							dev);

		goto done;
	}

	dev = g_new0(struct remote_dev_info, 1);

	bacpy(&dev->bdaddr, bdaddr);
	dev->class = class;
	if (name)
		dev->name = g_strdup(name);
	if (alias)
		dev->alias = g_strdup(alias);
	dev->legacy = legacy;
	dev->name_status = name_status;

	adapter->found_devices = g_slist_prepend(adapter->found_devices, dev);

done:
	dev->rssi = rssi;

	adapter->found_devices = g_slist_sort(adapter->found_devices,
						(GCompareFunc) dev_rssi_cmp);

	adapter_emit_device_found(adapter, dev);
}

int adapter_remove_found_device(struct btd_adapter *adapter, bdaddr_t *bdaddr)
{
	struct remote_dev_info *dev, match;

	memset(&match, 0, sizeof(struct remote_dev_info));
	bacpy(&match.bdaddr, bdaddr);

	dev = adapter_search_found_devices(adapter, &match);
	if (!dev)
		return -1;

	dev->name_status = NAME_NOT_REQUIRED;

	return 0;
}

void adapter_update_oor_devices(struct btd_adapter *adapter)
{
	GSList *l;

	for (l = adapter->oor_devices; l; l = l->next) {
		char address[18];
		const char *paddr = address;
		struct remote_dev_info *dev = l->data;

		ba2str(&dev->bdaddr, address);

		g_dbus_emit_signal(connection, adapter->path,
				ADAPTER_INTERFACE, "DeviceDisappeared",
				DBUS_TYPE_STRING, &paddr,
				DBUS_TYPE_INVALID);

		adapter->found_devices = g_slist_remove(adapter->found_devices, dev);
		dev_info_free(dev);
	}

	g_slist_free(adapter->oor_devices);
	adapter->oor_devices = NULL;

	adapter->oor_devices = g_slist_copy(adapter->found_devices);
}

void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
{
	const gchar *path = adapter_get_path(adapter);
	gboolean discoverable, pairable;
	uint8_t real_class[3];

	if (adapter->scan_mode == scan_mode)
		return;

	adapter_remove_discov_timeout(adapter);

	switch (scan_mode) {
	case SCAN_DISABLED:
		adapter->mode = MODE_OFF;
		discoverable = FALSE;
		pairable = FALSE;
		break;
	case SCAN_PAGE:
		adapter->mode = MODE_CONNECTABLE;
		discoverable = FALSE;
		pairable = adapter->pairable;
		break;
	case (SCAN_PAGE | SCAN_INQUIRY):
		adapter->mode = MODE_DISCOVERABLE;
		discoverable = TRUE;
		pairable = adapter->pairable;
		if (adapter->discov_timeout != 0)
			adapter_set_discov_timeout(adapter,
						adapter->discov_timeout);
		break;
	case SCAN_INQUIRY:
		/* Address the scenario where a low-level application like
		 * hciconfig changed the scan mode */
		if (adapter->discov_timeout != 0)
			adapter_set_discov_timeout(adapter,
						adapter->discov_timeout);

		/* ignore, this event should not be sent */
	default:
		/* ignore, reserved */
		return;
	}

	/* If page scanning gets toggled emit the Pairable property */
	if ((adapter->scan_mode & SCAN_PAGE) != (scan_mode & SCAN_PAGE))
		emit_property_changed(connection, adapter->path,
					ADAPTER_INTERFACE, "Pairable",
					DBUS_TYPE_BOOLEAN, &pairable);

	memcpy(real_class, adapter->dev.class, 3);
	if (adapter->svc_cache)
		real_class[2] = adapter->svc_cache;

	if (discoverable && adapter->pairable && adapter->discov_timeout > 0 &&
						adapter->discov_timeout <= 60)
		adapter_ops->set_limited_discoverable(adapter->dev_id,
							real_class, TRUE);
	else if (!discoverable)
		adapter_ops->set_limited_discoverable(adapter->dev_id,
							real_class, FALSE);

	emit_property_changed(connection, path,
				ADAPTER_INTERFACE, "Discoverable",
				DBUS_TYPE_BOOLEAN, &discoverable);

	adapter->scan_mode = scan_mode;
}

struct agent *adapter_get_agent(struct btd_adapter *adapter)
{
	if (!adapter || !adapter->agent)
		return NULL;

	return adapter->agent;
}

void adapter_add_connection(struct btd_adapter *adapter,
				struct btd_device *device, uint16_t handle)
{
	if (g_slist_find(adapter->connections, device)) {
		error("Unable to add connection %d", handle);
		return;
	}

	device_add_connection(device, connection, handle);

	adapter->connections = g_slist_append(adapter->connections, device);
}

void adapter_remove_connection(struct btd_adapter *adapter,
				struct btd_device *device, uint16_t handle)
{
	bdaddr_t bdaddr;

	if (!g_slist_find(adapter->connections, device)) {
		error("No matching connection for handle %u", handle);
		return;
	}

	device_remove_connection(device, connection, handle);

	adapter->connections = g_slist_remove(adapter->connections, device);

	/* clean pending HCI cmds */
	device_get_address(device, &bdaddr);
	hci_req_queue_remove(adapter->dev_id, &bdaddr);

	if (device_is_authenticating(device))
		device_cancel_authentication(device, TRUE);

	if (device_is_temporary(device)) {
		const char *path = device_get_path(device);

		debug("Removing temporary device %s", path);
		adapter_remove_device(connection, adapter, device);
	}
}

gboolean adapter_has_discov_sessions(struct btd_adapter *adapter)
{
	if (!adapter || !adapter->disc_sessions)
		return FALSE;

	return TRUE;
}

int btd_register_adapter_driver(struct btd_adapter_driver *driver)
{
	GSList *adapters;

	adapter_drivers = g_slist_append(adapter_drivers, driver);

	if (driver->probe == NULL)
		return 0;

	adapters = manager_get_adapters();
	g_slist_foreach(adapters, probe_driver, driver);

	return 0;
}

void btd_unregister_adapter_driver(struct btd_adapter_driver *driver)
{
	adapter_drivers = g_slist_remove(adapter_drivers, driver);
}

static void agent_auth_cb(struct agent *agent, DBusError *derr,
							void *user_data)
{
	struct service_auth *auth = user_data;

	device_set_authorizing(auth->device, FALSE);

	auth->cb(derr, auth->user_data);
}

static int btd_adapter_authorize(struct btd_adapter *adapter,
					const bdaddr_t *dst,
					const char *uuid,
					service_auth_cb cb, void *user_data)
{
	struct service_auth *auth;
	struct btd_device *device;
	struct agent *agent;
	char address[18];
	gboolean trusted;
	const gchar *dev_path;
	int err;

	ba2str(dst, address);
	device = adapter_find_device(adapter, address);
	if (!device)
		return -EPERM;

	/* Device connected? */
	if (!g_slist_find(adapter->connections, device))
		return -ENOTCONN;

	trusted = read_trust(&adapter->bdaddr, address, GLOBAL_TRUST);

	if (trusted) {
		cb(NULL, user_data);
		return 0;
	}

	device = adapter_find_device(adapter, address);
	if (!device)
		return -EPERM;

	agent = device_get_agent(device);

	if (!agent)
		agent = adapter->agent;

	if (!agent)
		return -EPERM;

	auth = g_try_new0(struct service_auth, 1);
	if (!auth)
		return -ENOMEM;

	auth->cb = cb;
	auth->user_data = user_data;
	auth->device = device;

	dev_path = device_get_path(device);

	err = agent_authorize(agent, dev_path, uuid, agent_auth_cb, auth, g_free);

	if (err == 0)
		device_set_authorizing(device, TRUE);

	return err;
}

int btd_request_authorization(const bdaddr_t *src, const bdaddr_t *dst,
		const char *uuid, service_auth_cb cb, void *user_data)
{
	struct btd_adapter *adapter;
	GSList *adapters;

	if (src == NULL || dst == NULL)
		return -EINVAL;

	if (bacmp(src, BDADDR_ANY) != 0)
		goto proceed;

	/* Handle request authorization for ANY adapter */
	adapters = manager_get_adapters();

	for (; adapters; adapters = adapters->next) {
		int err;
		adapter = adapters->data;

		err = btd_adapter_authorize(adapter, dst, uuid, cb, user_data);
		if (err == 0)
			return 0;
	}

	return -EPERM;

proceed:
	adapter = manager_find_adapter(src);
	if (!adapter)
		return -EPERM;

	return btd_adapter_authorize(adapter, dst, uuid, cb, user_data);
}

int btd_cancel_authorization(const bdaddr_t *src, const bdaddr_t *dst)
{
	struct btd_adapter *adapter = manager_find_adapter(src);
	struct btd_device *device;
	struct agent *agent;
	char address[18];
	int err;

	if (!adapter)
		return -EPERM;

	ba2str(dst, address);
	device = adapter_find_device(adapter, address);
	if (!device)
		return -EPERM;

	/*
	 * FIXME: Cancel fails if authorization is requested to adapter's
	 * agent and in the meanwhile CreatePairedDevice is called.
	 */

	agent = device_get_agent(device);

	if (!agent)
		agent = adapter->agent;

	if (!agent)
		return -EPERM;

	err = agent_cancel(agent);

	if (err == 0)
		device_set_authorizing(device, FALSE);

	return err;
}

static gchar *adapter_any_path = NULL;
static int adapter_any_refcount = 0;

const char *adapter_any_get_path(void)
{
	return adapter_any_path;
}

const char *btd_adapter_any_request_path(void)
{
	if (adapter_any_refcount > 0)
		return adapter_any_path;

	adapter_any_path = g_strdup_printf("%s/any", manager_get_base_path());
	adapter_any_refcount++;

	return adapter_any_path;
}

void btd_adapter_any_release_path(void)
{
	adapter_any_refcount--;

	if (adapter_any_refcount > 0)
		return;

	g_free(adapter_any_path);
	adapter_any_path = NULL;
}

gboolean adapter_is_pairable(struct btd_adapter *adapter)
{
	return adapter->pairable;
}

gboolean adapter_powering_down(struct btd_adapter *adapter)
{
	return adapter->off_requested;
}

int btd_adapter_restore_powered(struct btd_adapter *adapter)
{
	char mode[14], address[18];

	if (!adapter_ops)
		return -EINVAL;

	if (!main_opts.remember_powered)
		return -EINVAL;

	if (adapter->up)
		return 0;

	ba2str(&adapter->bdaddr, address);
	if (read_device_mode(address, mode, sizeof(mode)) == 0 &&
						g_str_equal(mode, "off"))
		return 0;

	return adapter_ops->set_powered(adapter->dev_id, TRUE);
}

int btd_adapter_switch_offline(struct btd_adapter *adapter)
{
	if (!adapter_ops)
		return -EINVAL;

	if (!adapter->up)
		return 0;

	return adapter_ops->set_powered(adapter->dev_id, FALSE);
}

int btd_register_adapter_ops(struct btd_adapter_ops *btd_adapter_ops)
{
	/* Already registered */
	if (adapter_ops)
		return -EALREADY;

	if (btd_adapter_ops->setup == NULL)
		return -EINVAL;

	adapter_ops = btd_adapter_ops;

	return 0;
}

void btd_adapter_cleanup_ops(struct btd_adapter_ops *btd_adapter_ops)
{
	adapter_ops->cleanup();
}

int adapter_ops_setup(void)
{
	if (!adapter_ops)
		return -EINVAL;

	return adapter_ops->setup();
}
